import pandas as pd
import ipywidgets as wd
from IPython.display import display, HTML, clear_output
import dictionary.data.access_data as dt
import os
import re


# styles
def get_styles():
    styles = {
        "table": "width: 100%; border-spacing: 0; border-bottom: 1px solid black;",
        "stats_header": "text-align: center; border-spacing: 0; border-bottom: 1px solid black; background-color: white; font-family: Helvetica, Neue; font-weight: bold; word-break:break-all;",
        "label": "text-align: left; font-weight: bold; font-size: 12px;  font-family: Helvetica, Neue",
        "helper": "text-align: left; font-size: 14px;margin-left: 5px; margin-right: 5px;font-family: Helvetica, Neue",
        "first_column": "background-color: #FFFFFF; text-align: left; width: 25%; font-size: 12px; font-weight: bold; font-family: Tahoma, Verdana, serif; border-bottom: 3px solid black;",
        "second_column": "background-color: #FFFFFF; text-align: left; width: 25%; font-size: 12px; font-family: Tahoma, Verdana, serif; border-bottom: 3px solid black;",
        "third_column": "background-color: #FFFFFF; text-align: right; font-size: 12px; font-family: Tahoma, Verdana, serif; border-bottom: 3px solid black;",
        "first_cell": "background-color: #FFFFFF; text-align: left; border-spacing: 0; border-bottom: 1px solid black;font-size: 12px; font-family:Tahoma, Verdana; word-break:break-all;",
        "second_cell": "background-color: #FFFFFF; text-align: right; border-spacing: 0;  border-bottom: 1px solid black;font-size: 12px; font-family: Tahoma, Verdana; word-break:break-all;"
    }
    return styles


style_base = {'font_size': '12px', 'text_color': 'black', 'background': 'rgb(247, 247, 247)'}
button_style = {'font_size': '12px', 'text_color': 'white', 'font_weight': 'bold',
                'font_family': ' Tahoma, Verdana,sans-serif', 'text_align': 'center'}

search_helper_text = f"""<p>This widget allows you to search for variables and descriptions across multiple datasets. Follow these instructions to effectively use the search functionality:
        <li><strong>If you want to search within a specific dataset:</strong> Use the "Datasets" dropdown menu to select the dataset you wish to search. Scroll down to see the list of available datasets and select your choice.
        </li><li><strong>If you want to search across all datasets:</strong>Leave the "Datasets" dropdown set to "None". This is the default option.
        </li>In the text box, type the word or phrase you want to search for and click the "Search" button. Note: You need to enter at least 3 characters for the search to function.</p>
        <p><strong>Search Results:</strong> The widget will display a table below the search box, showing the search results.
        <li><strong>If a specific dataset was selected:</strong> The table will show the variables and descriptions from that dataset that match your search terms.
        </li><li><strong>If "None" was selected:</strong> The table will show results from all datasets that match your search terms, including the dataset name, variable name, and description.
        </li></p><p><strong>Save Table:</strong> If you want to save the search results as an HTML file, click the "Save Table" button.
        <li><strong>If a specific dataset was selected:</strong> The file will be named using the dataset name and the search terms (e.g., "Food Security Data 2021_searchterm.html").
        </li><li><strong>If "None" was selected:</strong> The file will be named using "Datasets" and the search terms (e.g., "Datasets_with_searchterm.html").
        </li>A confirmation message will appear below the "Save Table" button, indicating the file name and location.</p>"""


def create_helper(text, helper_name, writing_style=None):
    if writing_style is None:
        writing_style = get_styles().get('helper', '')
    helper = wd.HTML(value=f"""<div style="{writing_style}"><p>{text}</p></div>""",
                     layout=wd.Layout(grid_area=f'{helper_name}_helper_box', border='1px solid gray', width='96%',
                                      justify='center'),
                     style={**style_base, 'word_break': 'break_all', 'padding': '3px'}, disabled=False)
    return helper


def create_select_dropdown(options=None, box_name=None, value=None):
    select_dropdown = wd.Select(
        options=options, disabled=False, rows=10, value=value,
        layout=wd.Layout(grid_area=f'{box_name}_select_box', width='100%',
                         background_color='white', border='1px solid #ababab'))
    return select_dropdown


def create_label(text, writing_style=None):
    lower_text = text.lower()
    label = wd.HTML(value=f"""<div style="{writing_style}"><p>{text}</p></div>""",
                    disabled=False,
                    layout=wd.Layout(grid_area=f'{lower_text}_label_box', width='100%', border='1px solid #ababab'),
                    style={**style_base, 'border': '1px solid #ababab'})
    return label


def create_button(text, box_name, style):
    button = wd.Button(description=f'{text}', style=style,
                       layout=wd.Layout(grid_area=f'{box_name}_button_box', width='100%'))
    return button


def search_data():
    dir_path = os.path.dirname(os.path.abspath(__file__))
    csv_path = os.path.join(dir_path, 'data', 'MainTableDatasets.csv')
    dataset_info = dt.get_data(csv_path)
    data_directory = os.path.join(dir_path, 'data')
    search_helper = create_helper(text=search_helper_text, helper_name='search')
    styles = get_styles()
    table_output = wd.Output()
    dataset_select = create_select_dropdown(options=["None"] + dataset_info['dataset_title'].tolist(),
                                            box_name='dataset', value='None')

    search_area = wd.Textarea(value='', placeholder='Type at least 3 characters',
                              layout=wd.Layout(grid_area='search_text_box', width='90%',
                                               background_color='white', border='1px solid #ababab'), disabled=False)

    search_button = create_button(text='Search', box_name='search', style={**button_style, 'button_color': 'blue'})
    save_button = create_button(text='Save Table', box_name='save', style={**button_style, 'button_color': 'green'})
    clear_button = create_button(text='Clear Table', box_name='clear', style={**button_style, 'button_color': 'red'})

    def generate_variable_table(df):
        html_content = f"""<html> <table style="{styles['table']}">
            <tr><th style="{styles['first_column']}">Variables</th>
            <th style="{styles['second_column']}">Descriptions</th>
            </tr>
        """
        for _, row in df.iterrows():
            html_content += f"""<tr><td style="{styles['first_cell']}">{row.iloc[0]}</td>
                <td style="{styles['second_cell']}">{row.iloc[1]}</td></tr>
            """
        html_content += "</table></html>"

        with open("variables.html", 'w') as file:
            file.write(html_content)
        return html_content

    def generate_combined_table(df):
        html_content = f"""
        <html><table style="{styles['table']}">
            <tr><th style="{styles['first_column']}">Dataset</th>
            <th style="{styles['second_column']}">Variables</th>
            <th style="{styles['third_column']}">Description</th>
            </tr>"""

        for _, row in df.iterrows():
            html_content += f"""
            <tr><td style="{styles['first_cell']}">{row.iloc[0]}</td>
            <td style="{styles['first_cell']}">{row.iloc[1]}</td>
            <td style="{styles['second_cell']}">{row.iloc[2]}</td>
            </tr>"""
        html_content += "</table></html>"

        with open("combined_table.html", 'w') as file:
            file.write(html_content)
        return html_content

    table_display = wd.HTML(value="")
    error_output = wd.Output()

    def update_table(search_text):
        pattern = ".*" + re.escape(search_text) + ".*" if search_text else ".*"

        if dataset_select.value == "None":
            data_list = []
            for idx, ds in dataset_info.iterrows():
                ds_title = ds['dataset_title']
                file_name = ds['file_name']
                file_path = os.path.join(data_directory, file_name)
                try:
                    df = pd.read_csv(file_path)
                    if 'Variable' in df.columns and 'Description' in df.columns:
                        temp_df = df[['Variable', 'Description']]
                        temp_df.columns = ['Variable Name', 'Description']

                        filtered_df = temp_df[
                            temp_df['Variable Name'].str.contains(pattern, case=False, na=False, regex=True) |
                            temp_df['Description'].str.contains(pattern, case=False, na=False, regex=True)
                            ]
                        if not filtered_df.empty:
                            filtered_df.insert(0, 'Dataset', ds_title)
                            data_list.append(filtered_df)
                    else:
                        print(f"Dataset {ds_title} does not have required columns.")
                except Exception as e:
                    print(f"Error loading dataset {ds_title}: {e}")
            if data_list:
                combined_df = pd.concat(data_list, ignore_index=True)
            else:
                combined_df = pd.DataFrame(columns=['Dataset', 'Variable Name', 'Description'])

            var_html = generate_combined_table(combined_df)
            table_display.value = var_html
        else:

            file_info = dataset_info.loc[dataset_info['dataset_title'] == dataset_select.value, 'file_name']
            if not file_info.empty:
                file_name = file_info.values[0]
                file_path = os.path.join(data_directory, file_name)

                try:
                    df = pd.read_csv(file_path)
                    if 'Variable' in df.columns and 'Description' in df.columns:
                        temp_df = df[['Variable', 'Description']].copy()
                        temp_df.columns = ['Variable Name', 'Description']

                        filtered = temp_df[
                            temp_df['Variable Name'].str.contains(pattern, case=False, na=False, regex=True) |
                            temp_df['Description'].str.contains(pattern, case=False, na=False, regex=True)
                            ]
                        if not filtered.empty:
                            var_html = generate_variable_table(filtered)
                            table_display.value = var_html
                        else:
                            table_display.value = "No matching records found."
                    else:
                        table_display.value = "Error: Dataset does not have required columns."
                except Exception as e:
                    table_display.value = f"Error loading dataset: {e}"
            else:
                table_display.value = "Error: Dataset not found in MainTableDatasets.csv."

    def on_dataset_change(change):
        if change['type'] == 'change' and change['name'] == 'value':
            update_table(search_area.value)

    dataset_select.observe(on_dataset_change, 'value')

    def on_search_clicked(b):
        error_output.clear_output()
        update_table(search_area.value)

    search_button.on_click(on_search_clicked)
    save_output = wd.Output(layout=wd.Layout(width='100%'))

    def save_filtered_table(_):
        try:
            if not table_display.value.strip():
                with save_output:
                    save_output.clear_output()
                    print("No data to save")
            else:
                if dataset_select.value == "None":
                    filename = f"Datasets_with_{search_area.value}.html"
                else:
                    filename = f"{dataset_select.value}_{search_area.value}.html"
                with open(filename, "w", encoding="utf-8") as file:
                    file.write(table_display.value)

                with save_output:
                    save_output.clear_output()
                    print(f"Table saved as '{filename}'.")

        except Exception as e:
            with save_output:
                save_output.clear_output()
                print(f"Error saving file: {e}")

    def clear_output(b):
        table_display.value = ""
        save_output.clear_output()
        error_output.clear_output()
        dataset_select.value = 'None'
        search_area.value = ''

    save_button.on_click(save_filtered_table)
    clear_button.on_click(clear_output)

    error_output.layout.grid_area = 'error_output_box'

    dataset_label = wd.Label(value='Datasets', layout=wd.Layout(grid_area='dataset_label_box'))
    search_label = wd.Label(value='Search', layout=wd.Layout(grid_area='search_label_box'))

    search_grid_layout = wd.GridBox(
        children=[search_helper, dataset_label, dataset_select,
                  search_area, search_label, search_button, save_output, save_button,
                  clear_button, error_output],
        layout=wd.Layout(grid_template_columns='40% 15% 30% 15%',
                         grid_template_rows='auto',
                         display='grid',
                         grid_template_areas='''
                         "search_helper_box search_helper_box  search_helper_box search_helper_box"
                        "dataset_label_box search_label_box . ."
                         "dataset_select_box  search_text_box  search_text_box search_text_box "
                        "dataset_select_box  . search_button_box ."
                         "dataset_select_box  . save_button_box ."
                          "dataset_select_box  . clear_button_box ."
                           "dataset_select_box  . error_output_box ."
                          "save_output_box  save_output_box save_output_box save_output_box"
                        ''',
                         grid_gap='5px',
                         width='98%'
                         ))
    display(wd.VBox([search_grid_layout, table_display]))


search_data()


from .create_labels import get_facts, get_meta, get_provenance, get_dictionary, get_data_summary, \
    get_summary_stats, get_show_barplots, get_show_pairplots, get_show_correlations

def labels():
    get_facts()
    get_meta()
    get_provenance()
    get_dictionary()
    get_data_summary()
    get_summary_stats()
    get_show_barplots()
    get_show_pairplots()
    get_show_correlations()

labels()



from pyschare.helpers.create_widgets import create_button, create_label, create_select_dropdown, get_layout, \
    create_multiple_select_dropdown, create_dropdown, create_helper
from pyschare.helpers.styles import get_styles
from pyschare.helpers.data_functions import get_main_table_path, get_data_dir, get_data, get_visual_table_info, \
    get_main_table_info, get_bucket, select_data_options, get_dropdown_value, get_input_path, save_data_to_bucket, \
    save_to_bucket, load_data, parse_data, get_parsed_data

from pyschare.helpers.constants import DATAPATH, MAIN_TABLE, MAIN_TITLE, MAIN_TABLE_VAR, VISUAL_TABLE, FILEPATH,  \
    visual_helper_text, search_helper_text, subset_helper_text, select_helper_text, calculate_helper_text, \
    data_explore_helper_text

__all__ = ['create_button', 'create_label', 'create_select_dropdown',
           'get_layout', 'create_multiple_select_dropdown', 'create_dropdown', 'create_helper', 'get_styles',
           'get_main_table_path', 'get_data_dir', 'get_data', 'get_visual_table_info',
           'get_main_table_info', 'get_bucket', 'select_data_options',
           'get_dropdown_value', 'get_input_path', 'save_data_to_bucket',
           'save_to_bucket', 'load_data', 'parse_data', 'get_parsed_data',
           'DATAPATH', 'MAIN_TABLE', 'MAIN_TITLE', 'MAIN_TABLE_VAR',
           'VISUAL_TABLE', 'FILEPATH', 'visual_helper_text', 'search_helper_text',
           'subset_helper_text', 'select_helper_text', 'calculate_helper_text', 'data_explore_helper_text'

           ]



# Run Gemini Chat
import base64
import json
import logging
import markdown2
import os
import requests
import subprocess
import time

# Google Cloud authentication
from google import genai
from google.genai import types as genai_types
from google.auth import default
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import Flow
from google.oauth2.credentials import Credentials as OAuth2Credentials

# Google Cloud services
from google.api_core.exceptions import GoogleAPICallError, PermissionDenied
from google.cloud import aiplatform, billing, resourcemanager_v3
from google.cloud.resourcemanager_v3 import types

# Vertex AI
import vertexai
from vertexai.generative_models import GenerativeModel, Part, SafetySetting

# IPython and Jupyter
from IPython.display import HTML, clear_output, display, Markdown
import ipywidgets as widgets

# Code highlighting
from pygments import highlight
from pygments.lexers import get_lexer_by_name, guess_lexer
from pygments.formatters import HtmlFormatter
from pygments.util import ClassNotFound

def get_gemini_chat():
    
    llm_project_id = "schare-gemini"

    # Define model options and their corresponding locations
    model_options = {
        "Gemini 2.0 Flash": {
            "model_name": "gemini-2.0-flash-001",
            "locations": ["us-east5","us-central1"]
        }
    }

    # System instructions
    system_instructions = """You are an advanced AI coding expert with extensive knowledge across multiple programming languages, frameworks, and software development practices. Your task is to provide high-quality, professional-grade assistance based on the given context and user's latest message.

    Instructions:
    1. Code Presentation:
      - Always present code within triple backtick blocks (```), specifying the language at the start.
      - Each language should have its own code block and avoid mixing languages. For example, instead of a single html app do a code block for html, css, and js separately. 
    2. Comprehensive Analysis:
      - Analyze the provided code files and chat history thoroughly before responding.
      - Consider architectural implications, potential optimizations, and best practices in your solutions.
    3. Advanced Problem-Solving:
      - Break down complex tasks into logical, manageable steps.
      - Provide algorithmic insights and time/space complexity analysis where relevant.
    4. Detailed Explanations:
      - Offer in-depth explanations using both inline comments and separate markdown sections.
      - Elucidate your reasoning, especially for architectural decisions or non-obvious optimizations.
    5. Best Practices and Optimizations:
      - Proactively suggest improvements, refactoring opportunities, and performance optimizations.
      - Incorporate modern development practices, design patterns, and industry standards in your solutions.
    6. Clarity and Precision:
      - If the user's request is ambiguous, ask targeted questions to clarify requirements before proceeding.
      - Anticipate potential issues or edge cases in your solutions and address them preemptively.
    7. Full File Requests:
       - If the user asks for a full file, return the entire file with any suggested upgrades or fixes.
       - Ensure that the returned file maintains the current logic and structure, only implementing necessary changes.
       - Present the full file in a single code block, preserving all original comments and formatting.
       - After the code block, provide a summary of the changes made and the rationale behind them.

    Provide a comprehensive, expert-level response that addresses the user's latest message while leveraging the full context provided. Your solution should reflect the depth and breadth of knowledge expected from a senior software engineer or technical lead."""

    # Custom CSS for styling
    custom_css = """
    <style>
        .output-container {
            border: 1px solid #ddd;
            border-radius: 5px;
            padding: 10px;
            margin: 10px 0;
            background-color: #f9f9f9;
        }
        .code-block {
            background-color: #f6f8fa;
            border-radius: 3px;
            padding: 10px;
            font-family: monospace;
            white-space: pre-wrap;
            overflow-x: auto;
        }
        .user-question {
            font-style: italic;
            color: #555;
            margin-bottom: 10px;
        }
        .model-info {
            font-weight: bold;
            color: #0066cc;
            margin-bottom: 10px;
        }
        .markdown-body {
            font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', 'Noto Sans', Helvetica, Arial, sans-serif, 'Apple Color Emoji', 'Segoe UI Emoji';
            font-size: 16px;
            line-height: 1.5;
            word-wrap: break-word;
        }
        .markdown-body h1, .markdown-body h2, .markdown-body h3, .markdown-body h4, .markdown-body h5, .markdown-body h6 {
            margin-top: 24px;
            margin-bottom: 16px;
            font-weight: 600;
            line-height: 1.25;
        }
        .markdown-body code {
            padding: 0.2em 0.4em;
            margin: 0;
            font-size: 85%;
            background-color: rgba(175, 184, 193, 0.2);
            border-radius: 6px;
        }
        .markdown-body pre {
            padding: 16px;
            overflow: auto;
            font-size: 85%;
            line-height: 1.45;
            background-color: #f6f8fa;
            border-radius: 6px;
        }
        .markdown-body pre code {
            display: inline;
            max-width: auto;
            padding: 0;
            margin: 0;
            overflow: visible;
            line-height: inherit;
            word-wrap: normal;
            background-color: transparent;
            border: 0;
        }
    </style>
    """

    # Function to format code with syntax highlighting
    def format_code(code, language=None):
        try:
            if language:
                lexer = get_lexer_by_name(language, stripall=True)
            else:
                lexer = guess_lexer(code)
            formatter = HtmlFormatter(style="friendly", linenos=False, cssclass="code-block")
            return highlight(code, lexer, formatter)
        except ClassNotFound:
            # If language detection fails, return plain text
            return f'<pre class="code-block">{code}</pre>'

    # Function to process and format the AI response
    def process_ai_response(response):
        parts = response.split("```")
        formatted_parts = []
        for i, part in enumerate(parts):
            if i % 2 == 0:
                # This is not a code block, render it as markdown
                html_content = markdown2.markdown(part, extras=["fenced-code-blocks", "tables"])
                formatted_parts.append(f'<div class="markdown-body">{html_content}</div>')
            else:
                # This is a code block
                lines = part.split("\n")
                if len(lines) > 1:
                    language = lines[0].strip()
                    code = "\n".join(lines[1:])
                    formatted_parts.append(format_code(code, language))
                else:
                    formatted_parts.append(format_code(part))
        return "".join(formatted_parts)

    # Create dropdowns
    model_dropdown = widgets.Dropdown(
        options=list(model_options.keys()),
        description='Select Model:',
        style={'description_width': 'initial'}
    )

    location_dropdown = widgets.Dropdown(
        description='Select Location:',
        style={'description_width': 'initial'}
    )

    # Function to update location dropdown based on selected model
    def update_locations(*args):
        model = model_dropdown.value
        location_dropdown.options = model_options[model]['locations']
        location_dropdown.value = model_options[model]['locations'][0]

    # Attach the update function to the model dropdown
    model_dropdown.observe(update_locations, names='value')

    # Initialize the location dropdown
    update_locations()

    # Generation configurations
    generation_config = {
        "max_output_tokens": 8192,
        "temperature": 0.7,
        "top_p": 0.95,
    }

    safety_settings = [
        SafetySetting(
            category=SafetySetting.HarmCategory.HARM_CATEGORY_HATE_SPEECH,
            threshold=SafetySetting.HarmBlockThreshold.BLOCK_NONE
        ),
        SafetySetting(
            category=SafetySetting.HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT,
            threshold=SafetySetting.HarmBlockThreshold.BLOCK_NONE
        ),
        SafetySetting(
            category=SafetySetting.HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT,
            threshold=SafetySetting.HarmBlockThreshold.BLOCK_NONE
        ),
        SafetySetting(
            category=SafetySetting.HarmCategory.HARM_CATEGORY_HARASSMENT,
            threshold=SafetySetting.HarmBlockThreshold.BLOCK_NONE
        ),
    ]

    # Function to generate content using Gemini 1.5
    def generate_gemini_1_5(user_input):
        access_token = get_access_token()
        
        credentials = OAuth2Credentials(
            token=access_token,
            refresh_token=None,
            token_uri=None,
            client_id=None,
            client_secret=None,
            scopes=None
        )

        # Initialize Vertex AI with credentials
        vertexai.init(
            project=llm_project_id, 
            location=location_dropdown.value,
            credentials=credentials
        )
        
        model = GenerativeModel(model_options[model_dropdown.value]['model_name'])

        # Combine system instructions and user input
        combined_input = f"""Instructions for AI Assistant:
    {system_instructions}

    User Question:
    {user_input}

    AI Assistant: Certainly! I'll provide a comprehensive, expert-level response to your question while following the given instructions. Let's begin:

    """

        responses = model.generate_content(
            [combined_input],
            generation_config=generation_config,
            safety_settings=safety_settings,
            stream=True,
        )

        full_response = ""
        for response in responses:
            full_response += response.text
        return full_response

    # Function to generate content using Gemini 2.0
    def generate_gemini_2_0(user_input):
        access_token = get_access_token()
        
        credentials = OAuth2Credentials(
            token=access_token,
            refresh_token=None,
            token_uri=None,
            client_id=None,
            client_secret=None,
            scopes=None
        )

        # Initialize Vertex AI with credentials
        vertexai.init(
            project=llm_project_id, 
            location=location_dropdown.value,
            credentials=credentials
        )
        
        model = GenerativeModel(model_options[model_dropdown.value]['model_name'])

        # Combine system instructions and user input
        combined_input = f"""Instructions for AI Assistant:
    {system_instructions}

    User Question:
    {user_input}

    AI Assistant: Certainly! I'll provide a comprehensive, expert-level response to your question while following the given instructions. Let's begin:

    """

        responses = model.generate_content(
            [combined_input],
            generation_config=generation_config,
            safety_settings=safety_settings,
            stream=True,
        )

        full_response = ""
        for response in responses:
            full_response += response.text
        return full_response

    def get_access_token():
        # Run the gcloud command and capture output
        cmd = "gcloud auth print-access-token --impersonate-service-account=gemini-only@schare-gemini.iam.gserviceaccount.com"
        result = subprocess.run(cmd.split(), capture_output=True, text=True)

        # Return the token (stripped of whitespace)
        return result.stdout.strip()

    # Update the run_generation function to handle errors
    def run_generation(b):
        with output_area:
            output_area.clear_output(wait=True)
            display(HTML(custom_css))
            display(HTML(f"""
            <div class="output-container">
                <div class="model-info">Generating with {model_dropdown.value} in {location_dropdown.value}</div>
                <div class="user-question">User Question: {user_input.value}</div>
            </div>
            """))

            try:
                if model_dropdown.value == "Gemini 1.5 Flash":
                    response = generate_gemini_1_5(user_input.value)
                elif model_dropdown.value == "Gemini 2.0 Flash":
                    response = generate_gemini_2_0(user_input.value)
                else:
                    response = generate_claude(user_input.value)

                formatted_response = process_ai_response(response)
                display(HTML(f"""
                <div class="output-container">
                    <h3>AI Response:</h3>
                    {formatted_response}
                </div>
                """))
            except Exception as e:
                logging.exception("An error occurred during generation")
                display(HTML(f"""
                <div class="output-container error">
                    <h3>Error:</h3>
                    <p>An error occurred during generation: {str(e)}</p>
                </div>
                """))

    # Create a text area for user input
    user_input = widgets.Textarea(
        value='',
        placeholder='Type your coding question here...',
        description='Question:',
        disabled=False,
        layout=widgets.Layout(width='100%', height='150px')
    )

    # Create an output widget
    output_area = widgets.Output()

    # Create a button to trigger generation
    generate_button = widgets.Button(description="Generate")
    generate_button.on_click(run_generation)

    # Display widgets
    display(model_dropdown, location_dropdown, user_input, generate_button, output_area)



import re
import pandas as pd

def process_tsv_file():
    df = pd.read_csv('A_MainTableDatasets.tsv', sep="\t")
    return df

def normalize_dataset_names():
    df = process_tsv_file()
    names = df['entity:A_MainTableDatasets_id']
    new_names = []
    normalize_text(names, new_names)
    df['main_table_dataset_id'] = new_names
    return df


def normalize_text(names, normalized_names):
    for name in names:
        year_match = re.search(r'(\d{4})', name)
        year = year_match.group(1) if year_match else ""
        name_without_year = re.sub(r'(\d{4})', '', name)
        name_split = re.sub(r'(?<!_)([A-Z][a-z])', r'_\1', name_without_year)
        name_split = re.sub(r'([a-z])([A-Z])', r'\1_\2', name_split)
        name_split = re.sub(r'[-\s]+', '_', name_split)
        name_split = re.sub(r'_+', '_', name_split).strip('_')
        normalized_name = f"{name_split}_{year}" if year else name_split
        normalized_names.append(normalized_name.lower())



import os
import pandas as pd
import re
import warnings
import ipywidgets as wd

warnings.filterwarnings("ignore")
from pyschare.helpers.constants import MAIN_TITLE, FILEPATH

from pyschare.helpers.styles import get_styles
from pyschare.helpers.data_functions import get_data_dir, get_main_table_info



styles = get_styles()
def generate_variable_table( df):
    html_content = f"""<html> <table style="{styles['table']}">
        <tr><th style="{styles['first_column']}">Variables</th>
        <th style="{styles['second_column']}">Descriptions</th>
        </tr>
    """
    for _, row in df.iterrows():
        html_content += f"""<tr><td style="{styles['first_cell']}">{row.iloc[0]}</td>
            <td style="{styles['second_cell']}">{row.iloc[1]}</td></tr>
        """
    html_content += "</table></html>"

    with open("variables.html", 'w') as file:
        file.write(html_content)
    return html_content


def generate_combined_table(df):
    html_content = f"""
    <html><table style="{styles['table']}">
        <tr><th style="{styles['first_column']}">Dataset</th>
        <th style="{styles['second_column']}">Variables</th>
        <th style="{styles['third_column']}">Description</th>
        </tr>"""

    for _, row in df.iterrows():
        html_content += f"""
        <tr><td style="{styles['first_cell']}">{row.iloc[0]}</td>
        <td style="{styles['first_cell']}">{row.iloc[1]}</td>
        <td style="{styles['second_cell']}">{row.iloc[2]}</td>
        </tr>"""
    html_content += "</table></html>"
    #
    # with open("combined_table.html", 'w', encoding='utf-8') as file:
    #     file.write(html_content)
    return html_content

class _GeminiDataSearch:
    def __init__(self):
        self.dataset_info = get_main_table_info()
        self.data_directory = get_data_dir()
        self.table_display = wd.HTML(value="")


    def update_table(self, search_text):
        error_messages = []
        result_html = None
        search_words = search_text.strip().split() if search_text and search_text.strip() else []
        data_list = []
        for idx, ds in self.dataset_info.iterrows():
            ds_title = ds[MAIN_TITLE]
            file_name = ds[FILEPATH]
            file_path = os.path.join(self.data_directory, file_name)

            try:
                df = pd.read_csv(str(file_path))
            except FileNotFoundError:
                error_messages.append(f"ERROR: File not found for '{ds_title}': {file_path}")
                continue
            except UnicodeDecodeError:
                try:
                    df = pd.read_csv(str(file_path), encoding='latin-1')
                    error_messages.append(f"WARNING: Read '{ds_title}' with latin-1 encoding.")
                except Exception as e_enc:
                    error_messages.append(
                        f"ERROR: Failed to read '{ds_title}' with multiple encodings: {e_enc}")
                    continue
            except pd.errors.EmptyDataError:
                error_messages.append(f"WARNING: File for '{ds_title}' is empty or unparseable.")
                continue
            except Exception as e_read:
                error_messages.append(f"ERROR: Reading file for '{ds_title}': {e_read}")
                continue

            df.columns = df.columns.str.strip()

            if not ('Variable' in df.columns and 'Description' in df.columns):
                error_messages.append(
                    f"WARNING: Dataset '{ds_title}' skipped - missing 'Variable' or 'Description' column.")
                continue

            temp_df = df[['Variable', 'Description']].copy()
            temp_df.columns = ['Variable Name', 'Description']

            try:
                temp_df['Variable Name'] = temp_df['Variable Name'].astype(str).fillna('')
                temp_df['Description'] = temp_df['Description'].astype(str).fillna('')
            except Exception as e_conv:
                error_messages.append(
                    f"WARNING: Could not convert columns to string for '{ds_title}': {e_conv}. Skipping filtering for this file.")
                continue

            if not search_words:
                filtered_df = temp_df
            else:
                final_mask = pd.Series([True] * len(temp_df), index=temp_df.index)
                for word in search_words:
                    escaped_word = re.escape(word)
                    word_mask = (
                            temp_df['Variable Name'].str.contains(escaped_word, case=False, na=False,
                                                                  regex=True) |
                            temp_df['Description'].str.contains(escaped_word, case=False, na=False,
                                                                regex=True)
                    )
                    final_mask &= word_mask
                filtered_df = temp_df[final_mask]

            if not filtered_df.empty:
                filtered_df.insert(0, 'Dataset', ds_title)
                data_list.append(filtered_df)

            if error_messages:
                print("Encountered issues while searching:")
                for msg in error_messages:
                    print(f"- {msg}")

        if data_list:
            try:
                combined_df = pd.concat(data_list, ignore_index=True)
                if not combined_df.empty:
                    var_html = generate_combined_table(combined_df)
                    self.table_display.value = var_html
                else:
                    self.table_display.value = "<p>No matching records found after combining.</p>"

            except Exception as e_concat:
                self.table_display.value = ""
                print(f"FATAL ERROR: Failed to combine results: {e_concat}")
        else:
            if not error_messages:
                self.table_display.value = "<p>No matching records found.</p>"
            else:
                self.table_display.value = "<p>Search incomplete due to errors. No matching records found in successfully processed files.</p>"

        return result_html, error_messages






import ipywidgets as wd
from IPython.display import display, clear_output, HTML
import seaborn as sns
from matplotlib import pyplot as plt
import pandas as pd
import warnings

warnings.filterwarnings("ignore")

from pyschare.helpers.constants import calculate_helper_text, MAIN_TITLE
from pyschare.helpers.create_widgets import create_label, create_select_dropdown, create_dropdown, create_button, \
    create_helper
from pyschare.helpers.data_functions import select_data_options, get_visual_table_info, get_parsed_data, \
    get_main_table_info
from pyschare.helpers.styles import button_style


class _AdjustedVisuals:
    def __init__(self):
        self.data_info = get_visual_table_info()
        self.data_options = select_data_options(self.data_info)
        self.dataset_label = create_label(text='Dataset')
        self.x_axis_label = create_label(text='X')
        self.y_axis_label = create_label(text='Y')

        self.hue_dropdown_label = create_label(text='Hue')
        self.style_dropdown_label = create_label(text='Style')
        self.size_dropdown_label = create_label(text='Size')
        self.col_dropdown_label = create_label(text='Column')
        self.row_dropdown_label = create_label(text='Row')
        self.multiple_dropdown_label = create_label(text='Layer')
        self.plot_dropdown_label = create_label(text='Plot')

        self.dataset_dropdown = create_select_dropdown(
            options=['None'] + self.data_options,
            box_name='dataset',
            value='None'
        )

        default_options = ['None']
        default_value = 'None'
        self.x_axis_dropdown = create_dropdown(name='x', options=default_options)
        self.y_axis_dropdown = create_dropdown(name='y', options=default_options)
        self.hue_dropdown = create_dropdown(name='hue', options=default_options)
        self.style_dropdown = create_dropdown(name='style', options=default_options)
        self.size_dropdown = create_dropdown(name='size', options=default_options)
        self.col_dropdown = create_dropdown(name='col', options=default_options)
        self.row_dropdown = create_dropdown(name='row', options=default_options)

        self.multiple_dropdown = create_dropdown(name='layer',
                                                 options=[('None', 'None'), ('Layer', 'layer'),
                                                          ('Dodge', 'dodge'),
                                                          ('Stack', 'stack'), ('Fill', 'fill')],
                                                 value='layer')  # Default 'layer'
        self.plot_dropdown = create_select_dropdown(
            options=[('None', 'None'), ('Bar Plot', 'bar'), ('Box Plot', 'box'), ('Boxen Plot', 'boxen'),
                     ('Count Plot', 'count'),
                     ('Histogram', 'histogram'), ('Line Plot', 'line'), ('Point Plot', 'point'),
                     ('Scatter Plot', 'scatter'),
                     ('Strip Plot', 'strip'), ('Swarm Plot', 'swarm'), ('Violin Plot', 'violin')],
            box_name='plot',
            value='None')

        self.dataset_dropdown.observe(self.update_variable_options, names='value')

        self.show_plot_button = create_button(text='Show Plot', box_name='show_plot',
                                              style={**button_style, 'button_color': 'blue'})

        self.clear_plot_button = create_button(text='Clear Output', box_name='clear_plot',
                                               style={**button_style, 'button_color': 'red'})

        self.visual_helper = create_helper(text=calculate_helper_text, helper_name='visual')

        self.show_plot_output = wd.Output(layout=wd.Layout(grid_area='output_box', width='98%', height='auto',
                                                           min_height='300px'))

        self.show_plot_button.on_click(self.show_plot)
        self.clear_plot_button.on_click(self.clear_output)

        self.visual_grid_layout = wd.GridBox(
            children=[self.visual_helper, self.dataset_label, self.dataset_dropdown,
                      self.x_axis_label, self.x_axis_dropdown,
                      self.y_axis_label, self.y_axis_dropdown,
                      self.hue_dropdown_label, self.hue_dropdown,
                      self.style_dropdown_label, self.style_dropdown,
                      self.size_dropdown_label, self.size_dropdown,
                      self.plot_dropdown_label, self.plot_dropdown,
                      self.col_dropdown_label, self.col_dropdown,
                      self.row_dropdown_label, self.row_dropdown,
                      self.multiple_dropdown_label, self.multiple_dropdown,
                      self.show_plot_button, self.clear_plot_button
                      ],
            layout=wd.Layout(display='grid',
                             grid_template_columns='15% 35% 15% 35%',
                             grid_template_rows='repeat(12, auto)',
                             grid_template_areas='''    
                                     "visual_helper_box visual_helper_box visual_helper_box visual_helper_box "
                                     "visual_helper_box visual_helper_box visual_helper_box visual_helper_box "
                                     "dataset_label_box  dataset_select_box x_label_box x_dropdown_box "
                                     " . dataset_select_box  y_label_box y_dropdown_box"
                                     " . dataset_select_box hue_label_box hue_dropdown_box"
                                     " . dataset_select_box  style_label_box style_dropdown_box" 
                                     "plot_label_box plot_select_box  size_label_box size_dropdown_box"
                                     " . plot_select_box  column_label_box col_dropdown_box "
                                     " . plot_select_box  row_label_box row_dropdown_box"
                                      " . plot_select_box layer_label_box layer_dropdown_box"
                                      ". . . ."
                                      " . show_plot_button_box . clear_plot_button_box"


                                      ''',
                             grid_gap='10px',
                             width='98%',
                             height='auto',
                             margin='5px 5px 5px 5px'

                             ))

        display(self.visual_grid_layout, self.show_plot_output)

    def show_plot(self, b):

        with self.show_plot_output:
            clear_output(wait=True)
            print("Generating plot...")
            self.create_plot()

    def update_variable_options(self, change):

        selected_title = change['new']
        default_options = ['None']
        default_value = 'None'

        if selected_title == 'None':
            self.x_axis_dropdown.options = default_options
            self.y_axis_dropdown.options = default_options
            self.hue_dropdown.options = default_options
            self.style_dropdown.options = default_options
            self.size_dropdown.options = default_options
            self.col_dropdown.options = default_options
            self.row_dropdown.options = default_options

            self.x_axis_dropdown.value = default_value
            self.y_axis_dropdown.value = default_value
            self.hue_dropdown.value = default_value
            self.style_dropdown.value = default_value
            self.size_dropdown.value = default_value
            self.col_dropdown.value = default_value
            self.row_dropdown.value = default_value
            return

        try:
            selected_data = self.data_info[self.data_info[MAIN_TITLE] == selected_title]

            if selected_data.empty:
                self.x_axis_dropdown.options = default_options
                self.x_axis_dropdown.value = default_value
                self.y_axis_dropdown.options = default_options
                self.y_axis_dropdown.value = default_value
                self.hue_dropdown.options = default_options
                self.hue_dropdown.value = default_value
                self.style_dropdown.options = default_options
                self.style_dropdown.value = default_value
                self.size_dropdown.options = default_options
                self.size_dropdown.value = default_value
                self.col_dropdown.options = default_options
                self.col_dropdown.value = default_value
                self.row_dropdown.options = default_options
                self.row_dropdown.value = default_value
                return

            dataset = get_parsed_data(selected_data)

            if dataset is not None and isinstance(dataset, pd.DataFrame):
                new_options = ['None'] + dataset.columns.tolist()

                self.x_axis_dropdown.options = new_options
                # self.x_axis_dropdown.value = default_value
                self.y_axis_dropdown.options = new_options
                # self.y_axis_dropdown.value = default_value
                self.hue_dropdown.options = new_options
                # self.hue_dropdown.value = default_value
                self.style_dropdown.options = new_options
                # self.style_dropdown.value = default_value
                self.size_dropdown.options = new_options
                # self.size_dropdown.value = default_value
                self.col_dropdown.options = new_options
                # self.col_dropdown.value = default_value
                self.row_dropdown.options = new_options
                # self.row_dropdown.value = default_value
            else:

                print(
                    f"Warning: get_parsed_data did not return a valid DataFrame for title '{selected_title}'. Resetting dependent options.")
                self.x_axis_dropdown.options = default_options
                # self.x_axis_dropdown.value = default_value
                self.y_axis_dropdown.options = default_options
                # self.y_axis_dropdown.value = default_value
                self.hue_dropdown.options = default_options
                # self.hue_dropdown.value = default_value
                self.style_dropdown.options = default_options
                # self.style_dropdown.value = default_value
                self.size_dropdown.options = default_options
                # self.size_dropdown.value = default_value
                self.col_dropdown.options = default_options
                # self.col_dropdown.value = default_value
                self.row_dropdown.options = default_options
                # self.row_dropdown.value = default_value

        except KeyError as e:

            self.x_axis_dropdown.options = default_options
            # self.x_axis_dropdown.value = default_value

        except Exception as e:

            self.x_axis_dropdown.options = default_options
            # self.x_axis_dropdown.value = default_value

    def get_x_axis(self):
        val = self.x_axis_dropdown.value
        return None if val == 'None' else val

    def get_y_axis(self):
        val = self.y_axis_dropdown.value
        return None if val == 'None' else val

    def get_hue(self):
        val = self.hue_dropdown.value
        return None if val == 'None' else val

    def get_style(self):
        val = self.style_dropdown.value
        return None if val == 'None' else val

    def get_size(self):
        val = self.size_dropdown.value
        return None if val == 'None' else val

    def get_multiple(self):
        return self.multiple_dropdown.value

    def get_column(self):
        val = self.col_dropdown.value
        return None if val == 'None' else val

    def get_row(self):
        val = self.row_dropdown.value
        return None if val == 'None' else val

    def get_plot(self):
        val = self.plot_dropdown.value
        return None if val == 'None' else val

    def create_plot(self):
        selected_title = self.dataset_dropdown.value
        selected_data = self.data_info[self.data_info[MAIN_TITLE] == selected_title]

        dataset = get_parsed_data(selected_data)
        if dataset is None:
            print("No dataset selected.")
            return

        num_x_categories = 1
        plot_aspect = 1
        plot_height = 4
        x_col = self.get_x_axis()
        if x_col and x_col in dataset.columns:
            num_x_categories = dataset[x_col].nunique()
            plot_aspect = 1 + num_x_categories * 0.05
            plot_height = 4

        plot_parameters = {
            'x': self.get_x_axis(),
            'data': dataset,
            'height': plot_height,  # new
            'aspect': plot_aspect  # new
        }
        temp_multiple = ""

        if self.get_y_axis() is not None:
            plot_parameters['y'] = self.get_y_axis()

        if self.get_hue() is not None:
            plot_parameters['hue'] = self.get_hue()

        if self.get_style() is not None:
            plot_parameters['style'] = self.get_style()
        if self.get_size() is not None:
            plot_parameters['size'] = self.get_size()

        if self.get_multiple() is not None:
            temp_multiple = self.get_multiple()

        if self.get_column() is not None:
            plot_parameters['col'] = self.get_column()
        if self.get_row() is not None:
            plot_parameters['row'] = self.get_row()

        plot_type = self.get_plot()

        with self.show_plot_output:
            clear_output(wait=True)
            if plot_type in ['bar', 'box', 'boxen', 'count', 'point', 'strip', 'swarm', 'violin']:
                g = sns.catplot(kind=plot_type, **plot_parameters)
                g.set_xticklabels(rotation=45, ha='right')
                # plt.tight_layout()
                plt.show()
                plt.close()

            elif plot_type == 'scatter':
                g = sns.relplot(kind='scatter', **plot_parameters)
                g.set_xticklabels(rotation=45, ha='right')
                # plt.tight_layout()
                plt.show()
                plt.close()

            elif plot_type == 'line':
                g = sns.relplot(kind='line', **plot_parameters)
                g.set_xticklabels(rotation=45, ha='right')
                # plt.tight_layout()
                plt.show()
                plt.close()


            elif plot_type == 'histogram':

                hist_params = plot_parameters.copy()
                is_univariate = (plot_parameters['x'] is not None and plot_parameters['y'] is None) or \
                                (plot_parameters['x'] is None and plot_parameters['y'] is not None)

                if is_univariate and temp_multiple:
                    hist_params['multiple'] = temp_multiple
                #                     print(f"  (Using multiple='{multiple_val}')")
                elif not is_univariate:
                    print("")

                g = sns.displot(kind='hist', **hist_params)
                plt.tight_layout()
                plt.show()
                plt.close()

    def clear_output(self, b):
        with self.show_plot_output:
            self.show_plot_output.clear_output(wait=True)
        self.dataset_dropdown.value = 'None'

        self.multiple_dropdown.value = 'layer'
        self.plot_dropdown.value = 'None'



import ipywidgets as wd
from IPython.display import display, clear_output, HTML
import numpy as np

from .constants import data_explore_helper_text, MAIN_TITLE
from .data_functions import (select_data_options, get_dropdown_value,
                                     get_parsed_data, get_bucket, save_to_bucket, get_main_table_info)
from .create_widgets import (create_select_dropdown,
                                                      create_multiple_select_dropdown, create_label, create_button,
                                                      create_helper)

from .styles import get_styles, button_style


def calculate_categorical(df):
    styles = get_styles()
    html_content = f"""<tr>
                <th style="{styles['first_header']}">type</th>
                <th style="{styles['first_header']}">name</th>
                <th style="{styles['header']}">count</th>
                <th style="{styles['header']}">missing</th>
                <th style="{styles['header']}">unique</th>
                <th style="{styles['header']}">mostFreq</th>
                <th style="{styles['header']}">leastFreq</th>"""
    categorical_cols = df.select_dtypes(exclude=[np.number]).columns
    for col in categorical_cols:
        html_content += f"""<tr><td style="{styles['cell']}">Categorical</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{col}</td>"""
        count = df[col].count()
        html_content += f"""<td style="{styles['stats_cell']}">{count}</td>"""
        missing = df[col].isnull().sum()
        missing_pct = (missing / len(df[col])) * 100
        html_content += f"""<td style="{styles['stats_cell']}">{missing_pct:.2f}%</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{df[col].nunique()}</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{df[col].mode()[0] if not df[col].mode().empty else 'N/A'}</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{df[col].value_counts().idxmin() if not df[col].value_counts().empty else 'N/A'}</td></tr>"""

    return html_content


def calculate_numeric(df):
    styles = get_styles()
    html_content = f"""<tr>
                <th style="{styles['first_header']}">type</th>
                <th style="{styles['first_header']}">name</th>
                <th style="{styles['header']}">count</th>
                <th style="{styles['header']}">missing</th>
                <th style="{styles['header']}">min</th>
                <th style="{styles['header']}">median</th>
                <th style="{styles['header']}">max</th>
                <th style="{styles['header']}">mean</th>
                <th style="{styles['header']}">stdDeviation</th>
                <th style="{styles['header']}">zeros</th>"""
    numeric_cols = df.select_dtypes(include=[np.number]).columns
    for col in numeric_cols:
        html_content += f"""<tr><td style="{styles['cell']}">Numeric</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{col}</td>"""
        count = df[col].count()
        html_content += f"""<td style="{styles['stats_cell']}">{count}</td>"""
        missing = df[col].isnull().sum()
        missing_pct = (missing / len(df[col])) * 100
        html_content += f"""<td style="{styles['stats_cell']}">{missing_pct:.2f}%</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{df[col].min():.2f}</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{df[col].median():.2f}</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{df[col].max():.2f}</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{df[col].mean():.2f}</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{df[col].std():.2f}</td>"""
        html_content += f"""<td style="{styles['stats_cell']}">{missing_pct:.2f}%</td>"""
    return html_content


class _Explore:
    def __init__(self):
        self.data_info = get_main_table_info()
        self.data_options = select_data_options(self.data_info)
        self.dataset_dropdown = create_select_dropdown(options=self.data_options, box_name='dataset')
        self.variable_dropdown = create_multiple_select_dropdown(options=[], box_name='variable')
        self.dataset_label = create_label(text='Dataset')
        self.variable_label = create_label(text='Variable')
        self.show_data_button = create_button(text='Show Data', box_name='show',
                                              style={**button_style, 'button_color': 'blue'})
        self.save_data_button = create_button(text='Save Data', box_name='save',
                                              style={**button_style, 'button_color': 'green'})
        self.clear_data_button = create_button(text='Clear Output', box_name='clear',
                                               style={**button_style, 'button_color': 'red'})
        self.describe_data_button = create_button(text='Describe Data', box_name='describe',
                                                  style={**button_style, 'button_color': 'darkblue'})
        self.describe_data_button.on_click(self.calculate_descriptive_stats)

        self.subset_helper = create_helper(text=data_explore_helper_text, helper_name='subset')
        self.show_data_output = wd.Output()
        self.describe_data_output = wd.Output()
        self.save_data_output = wd.Output(layout=wd.Layout(grid_area='save_output_box'))
        self.save_data_button.on_click(self.save_data)
        self.show_data_button.on_click(self.show_data)
        self.clear_data_button.on_click(self.clear_output)

        self.subset_grid_layout = wd.GridBox(
            children=[self.subset_helper, self.dataset_label, self.dataset_dropdown,
                      self.variable_dropdown, self.variable_label,
                      self.clear_data_button, self.describe_data_button, self.show_data_button, self.save_data_button,
                      self.save_data_output],
            layout=wd.Layout(display='grid',
                             grid_template_columns='40% 40% 20%',
                             grid_template_rows='repeat(8,auto)',
                             grid_template_areas='''
                                             "subset_helper_box subset_helper_box subset_helper_box"
                                             "subset_helper_box subset_helper_box  subset_helper_box"
                                              "dataset_label_box variable_label_box . "
                                              "dataset_select_box  variable_select_box clear_button_box"
                                              "dataset_select_box  variable_select_box   show_button_box"
                                              "dataset_select_box  variable_select_box   describe_button_box"
                                              "dataset_select_box  variable_select_box   save_button_box"
                                              " save_output_box save_output_box save_output_box"
                                           ''',

                             grid_gap='10px',
                             width='98%',
                             height='auto',
                             margin='5px',
                             overflow='hidden'))

        display(wd.VBox([self.subset_grid_layout, self.show_data_output, self.describe_data_output]))

        self.dataset_dropdown.observe(self.update_variable_options, names='value')

        self.input_data = self.data_info[self.data_info[MAIN_TITLE] == self.dataset_dropdown.value]

    def update_variable_options(self, change):
        input_data = self.get_dataset()

        if input_data is not None:
            self.variable_dropdown.options = input_data.columns.tolist()
            self.variable_dropdown.disabled = False
        else:
            self.variable_dropdown.options = []
            self.variable_dropdown.disabled = True

    #     def get_input(self):
    #         return self.dataset_dropdown.value

    def get_dataset(self):
        input_df = get_dropdown_value(self.dataset_dropdown)
        if input_df is not None:
            selected_data = self.data_info[self.data_info[MAIN_TITLE] == input_df]
            parsed_data = get_parsed_data(selected_data)
            return parsed_data
        else:
            return None

    def get_subset_data(self):
        parsed_data = self.get_dataset()
        #         parsed_data = new_data.to_csv(index= False)
        selected_variables = get_dropdown_value(self.variable_dropdown)
        if selected_variables:
            subset_data = parsed_data[list(selected_variables)]
        else:
            subset_data = parsed_data
        return subset_data

    def show_data(self, b):
        subset_data = self.get_subset_data()
        with self.show_data_output:
            self.show_data_output.clear_output(wait=True)
            if subset_data is not None:
                display(subset_data.head())
            else:
                print("No dataset selected. Please select a dataset.")

    def save_data(self, b):
        input_df = get_dropdown_value(self.dataset_dropdown)
        selected_data = self.data_info[self.data_info[MAIN_TITLE] == input_df]
        parsed_data = self.get_subset_data()
        if parsed_data is not None:
            my_bucket = get_bucket()
            dataset_name = selected_data.iloc[0]['file_name']
            bucket_path = f"{my_bucket}/subset_{dataset_name}"
            try:
                save_to_bucket(parsed_data, bucket_path)

                with self.save_data_output:
                    self.save_data_output.clear_output()
                    print(f"Saving data to: {bucket_path}")

            except Exception as e:
                with self.save_data_output:
                    self.save_data_output.clear_output()
                    print(f"Error saving data: {e}")

        else:
            with self.save_data_output:
                self.save_data_output.clear_output()
                print("No data to save. Please select a dataset first.")

    def calculate_descriptive_stats(self, b):
        dataset = self.get_subset_data()
        if dataset is not None:
            html_content = self.generate_html_stats(dataset)
            with self.describe_data_output:
                clear_output(wait=True)
                display(HTML(html_content))
        else:
            print("Failed to load data. Please check the dataset or file type.")

    def generate_html_stats(self, df):
        if df.empty:
            return "No data available."

        styles = get_styles()
        html_content = f"""<html><table style="{styles['table']}">"""
        html_content += f"""<tr><th colspan='10' style="{styles['title']}">Descriptive Statistics</th></tr>"""

        html_content += calculate_categorical(df)
        html_content += "</table><table>"
        html_content += calculate_numeric(df)
        html_content += "</table></html>"
        return html_content

    def clear_output(self, b):
        with self.show_data_output and self.save_data_output and self.describe_data_output:
            self.show_data_output.clear_output()
            self.save_data_output.clear_output()
            self.describe_data_output.clear_output()

        self.dataset_dropdown.value = 'None'
        self.variable_dropdown.value = []






import ipywidgets as wd
from IPython.display import display, clear_output
import seaborn as sns
from matplotlib import pyplot as plt


from .constants import calculate_helper_text, MAIN_TITLE
from .create_widgets import create_label, create_select_dropdown, create_dropdown, create_button, \
    create_helper
from .data_functions import select_data_options, get_parsed_data, \
    get_main_table_info
from .styles import button_style

class _Visuals:
    def __init__(self):
        self.data_info = get_main_table_info()
        self.data_options = select_data_options(self.data_info)
        self.dataset_label = create_label(text='Dataset')
        self.x_axis_label = create_label(text='X')
        self.y_axis_label = create_label(text='Y')

        self.hue_dropdown_label = create_label(text='Hue')
        self.style_dropdown_label = create_label(text='Style')
        self.size_dropdown_label = create_label(text='Size')
        self.col_dropdown_label = create_label(text='Column')
        self.row_dropdown_label = create_label(text='Row')
        self.multiple_dropdown_label = create_label(text='Layer')
        self.plot_dropdown_label = create_label(text='Plot')
        self.dataset_dropdown = create_select_dropdown(options=['None'] + self.data_options, box_name='dataset')

        self.x_axis_dropdown = create_dropdown(name='x', options=[])
        self.y_axis_dropdown = create_dropdown(name='y', options=[])
        self.hue_dropdown = create_dropdown(name='hue', options=[])
        self.style_dropdown = create_dropdown(name='style', options=[])
        self.size_dropdown = create_dropdown(name='size', options=[])

        self.col_dropdown = create_dropdown(name='col', options=[])
        self.row_dropdown = create_dropdown(name='row', options=[])

        self.multiple_dropdown = create_dropdown(name='layer',
                                                 options=[('None', 'None'), ('Layer', 'layer'), ('Dodge', 'dodge'),
                                                          ('Stack', 'stack'), ('Fill', 'fill')], value='layer')
        self.plot_dropdown = create_select_dropdown(
            options=[('None', 'None'), ('Bar Plot', 'bar'), ('Box Plot', 'box'), ('Boxen Plot', 'boxen'),
                     ('Count Plot', 'count'),
                     ('Histogram', 'histogram'), ('Line Plot', 'line'), ('Point Plot', 'point'),
                     ('Scatter Plot', 'scatter'),
                     ('Strip Plot', 'strip'), ('Swarm Plot', 'swarm'), ('Violin Plot', 'violin')], box_name='plot',
            value='None')

        self.dataset_dropdown.observe(self.update_variable_options, names='value')

        self.show_plot_button = create_button(text='Show Plot', box_name='show_plot',
                                              style={**button_style, 'button_color': 'blue'})

        self.clear_plot_button = create_button(text='Clear Output', box_name='clear_plot',
                                               style={**button_style, 'button_color': 'red'})

        self.visual_helper = create_helper(text=calculate_helper_text, helper_name='visual')
        self.show_plot_output = wd.Output(layout=wd.Layout(grid_area='output_box', width='98%', height='80vh'))

        self.show_plot_button.on_click(self.show_plot)
        self.clear_plot_button.on_click(self.clear_output)

        self.visual_grid_layout = wd.GridBox(
            children=[self.visual_helper, self.dataset_label, self.dataset_dropdown,
                      self.x_axis_label, self.x_axis_dropdown,
                      self.y_axis_label, self.y_axis_dropdown,
                      self.hue_dropdown_label, self.hue_dropdown,
                      self.style_dropdown_label, self.style_dropdown,
                      self.size_dropdown_label, self.size_dropdown,
                      self.plot_dropdown_label, self.plot_dropdown,
                      self.col_dropdown_label, self.col_dropdown,
                      self.row_dropdown_label, self.row_dropdown,
                      self.multiple_dropdown_label, self.multiple_dropdown,
                      self.show_plot_button, self.clear_plot_button
                      ],
            layout=wd.Layout(display='grid',
                             grid_template_columns='15% 35% 15% 35%',
                             grid_template_rows='repeat(12, auto)',
                             grid_template_areas='''    
                             "visual_helper_box visual_helper_box visual_helper_box visual_helper_box "
                             "visual_helper_box visual_helper_box visual_helper_box visual_helper_box "
                             "dataset_label_box  dataset_select_box x_label_box x_dropdown_box "
                             " . dataset_select_box  y_label_box y_dropdown_box"
                             " . dataset_select_box hue_label_box hue_dropdown_box"
                             " . dataset_select_box  style_label_box style_dropdown_box" 
                             "plot_label_box plot_select_box  size_label_box size_dropdown_box"
                             " . plot_select_box  column_label_box col_dropdown_box "
                             " . plot_select_box  row_label_box row_dropdown_box"
                              " . plot_select_box layer_label_box layer_dropdown_box"
                              ". . . ."
                              " . show_plot_button_box . clear_plot_button_box"


                              ''',
                             grid_gap='10px',
                             width='98%',
                             height='auto',
                             margin='5px 5px 5px 5px'

                             ))

        display(wd.VBox([self.visual_grid_layout, self.show_plot_output]))

        self.input_data = self.data_info[self.data_info[MAIN_TITLE] == self.dataset_dropdown.value]

    def show_plot(self, b):
        with self.show_plot_output:
            clear_output(wait=True)
            self.create_plot()

    def update_variable_options(self, change):
        selected_title = self.dataset_dropdown.value
        selected_data = self.data_info[self.data_info[MAIN_TITLE] == selected_title]
        dataset = get_parsed_data(selected_data)
        if dataset is not None:
            self.x_axis_dropdown.options = ['None'] + dataset.columns.tolist()
            self.y_axis_dropdown.options = ['None'] + dataset.columns.tolist()
            self.hue_dropdown.options = ['None'] + dataset.columns.tolist()
            self.style_dropdown.options = ['None'] + dataset.columns.tolist()
            self.size_dropdown.options = ['None'] + dataset.columns.tolist()
            self.col_dropdown.options = ['None'] + dataset.columns.tolist()
            self.row_dropdown.options = ['None'] + dataset.columns.tolist()

    def get_x_axis(self):
        return self.x_axis_dropdown.value

    def get_y_axis(self):
        return self.y_axis_dropdown.value

    def get_hue(self):
        return self.hue_dropdown.value

    def get_style(self):
        return self.style_dropdown.value

    def get_size(self):
        return self.style_dropdown.value

    def get_multiple(self):
        return self.multiple_dropdown.value

    def get_column(self):
        return self.col_dropdown.value

    def get_row(self):
        return self.row_dropdown.value

    def get_plot(self):
        return self.plot_dropdown.value

    def create_plot(self):
        selected_title = self.dataset_dropdown.value
        selected_data = self.data_info[self.data_info[MAIN_TITLE] == selected_title]

        dataset = get_parsed_data(selected_data)
        if dataset is None:
            print("No dataset selected.")
            return

        plot_parameters = {
            'x': self.get_x_axis(),
            'data': dataset
        }
        temp_multiple = ""

        if self.get_y_axis() is not None:
            plot_parameters['y'] = self.get_y_axis()

        if self.get_hue() is not None:
            plot_parameters['hue'] = self.get_hue()

        if self.get_style() is not None:
            plot_parameters['style'] = self.get_style()
        if self.get_size() is not None:
            plot_parameters['size'] = self.get_size()
        if self.get_multiple() is not None:
            temp_multiple = self.get_multiple()

        if self.get_column() is not None:
            plot_parameters['col'] = self.get_column()
        if self.get_row() is not None:
            plot_parameters['row'] = self.get_row()

        plot_type = self.get_plot()

        with self.show_plot_output:
            clear_output(wait=True)
            if plot_type in ['bar', 'box', 'boxen', 'count', 'point', 'strip', 'swarm', 'violin']:
                g = sns.catplot(kind=plot_type, **plot_parameters)
                g.set_xticklabels(rotation=45, ha='right')
                #                 plt.tight_layout()
                plt.show()
                plt.close()

            elif plot_type == 'scatter':
                g = sns.relplot(kind='scatter', **plot_parameters)
                g.set_xticklabels(rotation=45, ha='right')
                #                 plt.tight_layout()
                plt.show()
                plt.close()

            elif plot_type == 'line':
                g = sns.relplot(kind='line', **plot_parameters)
                g.set_xticklabels(rotation=45, ha='right')
                #                 plt.tight_layout()
                plt.show()
                plt.close()

            elif plot_type == 'histogram':
                g = sns.displot(kind='hist', multiple=temp_multiple, **plot_parameters)
                g.set_xticklabels(rotation=45, ha='right')
                #                 plt.tight_layout()
                plt.show()
                plt.close()

    def clear_output(self, b):
        with self.show_plot_output:
            self.show_plot_output.clear_output()
        self.dataset_dropdown.value = 'None'
        self.x_axis_dropdown.value = []
        self.y_axis_dropdown.value = []
        self.hue_dropdown.value = []
        self.style_dropdown.value = []
        self.size_dropdown.value = []

        self.col_dropdown.value = []
        self.row_dropdown.value = []

        self.multiple_dropdown.value = 'layer'
        self.plot_dropdown.value = 'None'




import base64
import json
import logging
import markdown2
import os
import requests
import subprocess
import time

# Google Cloud authentication
from google import genai
from google.genai import types as genai_types
from google.auth import default
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import Flow
from google.oauth2.credentials import Credentials as OAuth2Credentials

# Google Cloud services
from google.api_core.exceptions import GoogleAPICallError, PermissionDenied
from google.cloud import aiplatform, billing, resourcemanager_v3
from google.cloud.resourcemanager_v3 import types

# Vertex AI
import vertexai
from vertexai.generative_models import GenerativeModel, Part, SafetySetting

# IPython and Jupyter
from IPython.display import HTML, clear_output, display, Markdown
import ipywidgets as widgets

# Code highlighting
from pygments import highlight
from pygments.lexers import get_lexer_by_name, guess_lexer
from pygments.formatters import HtmlFormatter
from pygments.util import ClassNotFound
import pandas as pd

def get_system_instructions():
    system_instructions = """You are an advanced AI coding expert with extensive knowledge across multiple programming languages, frameworks, and software development practices. Your task is to provide high-quality, professional-grade assistance based on the given context and user's latest message.
        
        Instructions:
        1. Code Presentation:
          - Always present code within triple backtick blocks (```), specifying the language at the start.
          - Each language should have its own code block and avoid mixing languages. For example, instead of a single html app do a code block for html, css, and js separately. 
        2. Comprehensive Analysis:
          - Analyze the provided code files and chat history thoroughly before responding.
          - Consider architectural implications, potential optimizations, and best practices in your solutions.
        3. Advanced Problem-Solving:
          - Break down complex tasks into logical, manageable steps.
          - Provide algorithmic insights and time/space complexity analysis where relevant.
        4. Detailed Explanations:
          - Offer in-depth explanations using both inline comments and separate markdown sections.
          - Elucidate your reasoning, especially for architectural decisions or non-obvious optimizations.
        5. Best Practices and Optimizations:
          - Proactively suggest improvements, refactoring opportunities, and performance optimizations.
          - Incorporate modern development practices, design patterns, and industry standards in your solutions.
        6. Clarity and Precision:
          - If the user's request is ambiguous, ask targeted questions to clarify requirements before proceeding.
          - Anticipate potential issues or edge cases in your solutions and address them preemptively.
        7. Full File Requests:
           - If the user asks for a full file, return the entire file with any suggested upgrades or fixes.
           - Ensure that the returned file maintains the current logic and structure, only implementing necessary changes.
           - Present the full file in a single code block, preserving all original comments and formatting.
           - After the code block, provide a summary of the changes made and the rationale behind them.
        
        Provide a comprehensive, expert-level response that addresses the user's latest message while leveraging the full context provided. Your solution should reflect the depth and breadth of knowledge expected from a senior software engineer or technical lead."""
    return system_instructions

def get_custom_css():
    custom_css = """
       <style>
           .output-container {
               border: 1px solid #ddd;
               border-radius: 5px;
               padding: 10px;
               margin: 10px 0;
               background-color: #f9f9f9;
           }
           .code-block {
               background-color: #f6f8fa;
               border-radius: 3px;
               padding: 10px;
               font-family: monospace;
               white-space: pre-wrap;
               overflow-x: auto;
           }
           .user-question {
               font-style: italic;
               color: #555;
               margin-bottom: 10px;
           }
           .model-info {
               font-weight: bold;
               color: #0066cc;
               margin-bottom: 10px;
           }
           .markdown-body {
               font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', 'Noto Sans', Helvetica, Arial, sans-serif, 'Apple Color Emoji', 'Segoe UI Emoji';
               font-size: 16px;
               line-height: 1.5;
               word-wrap: break-word;
           }
           .markdown-body h1, .markdown-body h2, .markdown-body h3, .markdown-body h4, .markdown-body h5, .markdown-body h6 {
               margin-top: 24px;
               margin-bottom: 16px;
               font-weight: 600;
               line-height: 1.25;
           }
           .markdown-body code {
               padding: 0.2em 0.4em;
               margin: 0;
               font-size: 85%;
               background-color: rgba(175, 184, 193, 0.2);
               border-radius: 6px;
           }
           .markdown-body pre {
               padding: 16px;
               overflow: auto;
               font-size: 85%;
               line-height: 1.45;
               background-color: #f6f8fa;
               border-radius: 6px;
           }
           .markdown-body pre code {
               display: inline;
               max-width: auto;
               padding: 0;
               margin: 0;
               overflow: visible;
               line-height: inherit;
               word-wrap: normal;
               background-color: transparent;
               border: 0;
           }
       </style>
       """
    return custom_css

def format_code(code, language=None):
    try:
        if language:
            lexer = get_lexer_by_name(language, stripall=True)
        else:
            lexer = guess_lexer(code)
        formatter = HtmlFormatter(style="friendly", linenos=False, cssclass="code-block")
        return highlight(code, lexer, formatter)
    except ClassNotFound:
        # If language detection fails, return plain text
        return f'<pre class="code-block">{code}</pre>'

def process_ai_response(response):
    parts = response.split("```")
    formatted_parts = []
    for i, part in enumerate(parts):
        if i % 2 == 0:
            # This is not a code block, render it as markdown
            html_content = markdown2.markdown(part, extras=["fenced-code-blocks", "tables"])
            formatted_parts.append(f'<div class="markdown-body">{html_content}</div>')
        else:
            # This is a code block
            lines = part.split("\n")
            if len(lines) > 1:
                language = lines[0].strip()
                code = "\n".join(lines[1:])
                formatted_parts.append(format_code(code, language))
            else:
                formatted_parts.append(format_code(part))
    return "".join(formatted_parts)


# two or more words search
import os
import ipywidgets as wd
import pandas as pd
import re
from IPython.display import display, clear_output, HTML
import warnings

warnings.filterwarnings("ignore")
from pyschare.helpers.constants import search_helper_text, MAIN_TITLE, FILEPATH, DATAPATH, DICTPATH
from pyschare.helpers.create_widgets import create_helper, create_select_dropdown, create_button
from pyschare.helpers.styles import get_styles, button_style
from pyschare.helpers.data_functions import get_data_dir, get_main_table_info, get_dictionary_table_info, \
    get_visual_table_info


class _DataSearch:
    def __init__(self):
        # self.dir_path = os.path.dirname(__file__)
        # self.csv_path = get_main_table_path()
        # self.dataset_info = get_search_table_info()
        self.dataset_info = get_main_table_info()
        self.data_directory = get_data_dir()
        self.search_helper = create_helper(text=search_helper_text, helper_name='search')
        self.styles = get_styles()

        self.dataset_select = create_select_dropdown(options=["None"] + self.dataset_info[MAIN_TITLE].tolist(),
                                                     box_name='dataset', value='None')

        self.search_area = wd.Textarea(value='', placeholder='Type at least 3 characters',
                                       layout=wd.Layout(grid_area='search_text_box', width='90%',
                                                        background_color='white', border='1px solid #ababab'),
                                       disabled=False)

        self.search_button = create_button(text='Search', box_name='search',
                                           style={**button_style, 'button_color': 'blue'})
        self.save_button = create_button(text='Save Table', box_name='save',
                                         style={**button_style, 'button_color': 'green'})
        self.clear_button = create_button(text='Clear Table', box_name='clear',
                                          style={**button_style, 'button_color': 'red'})

        self.dataset_label = wd.Label(value='Datasets', layout=wd.Layout(grid_area='dataset_label_box'))
        self.search_label = wd.Label(value='Search', layout=wd.Layout(grid_area='search_label_box'))
        self.table_display = wd.HTML(value="")
        self.error_output = wd.Output()

        self.save_button.on_click(self.save_filtered_table)
        self.clear_button.on_click(self.clear_output)
        self.search_button.on_click(self.on_search_clicked)
        self.save_output = wd.Output(layout=wd.Layout(width='100%'))
        self.dataset_select.observe(self.on_dataset_change, 'value')

        self.error_output.layout.grid_area = 'error_output_box'

        self.search_grid_layout = wd.GridBox(
            children=[self.search_helper, self.dataset_label, self.dataset_select,
                      self.search_area, self.search_label, self.search_button, self.save_output, self.save_button,
                      self.clear_button, self.error_output],
            layout=wd.Layout(grid_template_columns='40% 15% 30% 15%',
                             grid_template_rows='auto',
                             display='grid',
                             grid_template_areas='''
                                     "search_helper_box search_helper_box  search_helper_box search_helper_box"
                                    "dataset_label_box search_label_box . ."
                                     "dataset_select_box  search_text_box  search_text_box search_text_box "
                                    "dataset_select_box  . search_button_box ."
                                     "dataset_select_box  . save_button_box ."
                                      "dataset_select_box  . clear_button_box ."
                                       "dataset_select_box  . error_output_box ."
                                      "save_output_box  save_output_box save_output_box save_output_box"
                                    ''',
                             grid_gap='5px',
                             width='98%'
                             ))
        display(wd.VBox([self.search_grid_layout, self.table_display]))

    def generate_variable_table(self, df):
        html_content = f"""<html> <table style="{self.styles['table']}">
            <tr><th style="{self.styles['first_column']}">Variables</th>
            <th style="{self.styles['second_column']}">Descriptions</th>
            </tr>
        """
        for _, row in df.iterrows():
            html_content += f"""<tr><td style="{self.styles['first_cell']}">{row.iloc[0]}</td>
                <td style="{self.styles['second_cell']}">{row.iloc[1]}</td></tr>
            """
        html_content += "</table></html>"

        with open("variables.html", 'w') as file:
            file.write(html_content)
        return html_content

    def generate_combined_table(self, df):
        html_content = f"""
        <html><table style="{self.styles['table']}">
            <tr><th style="{self.styles['first_column']}">Dataset</th>
            <th style="{self.styles['second_column']}">Variables</th>
            <th style="{self.styles['third_column']}">Description</th>
            </tr>"""

        for _, row in df.iterrows():
            html_content += f"""
            <tr><td style="{self.styles['first_cell']}">{row.iloc[0]}</td>
            <td style="{self.styles['first_cell']}">{row.iloc[1]}</td>
            <td style="{self.styles['second_cell']}">{row.iloc[2]}</td>
            </tr>"""
        html_content += "</table></html>"

        with open("combined_table.html", 'w', encoding='utf-8') as file:
            file.write(html_content)
        return html_content

    def update_table(self, search_text):
        self.error_output.clear_output()
        error_messages = []

        search_words = search_text.strip().split() if search_text and search_text.strip() else []


        if self.dataset_select.value == "None":
            data_list = []
            for idx, ds in self.dataset_info.iterrows():
                ds_title = ds[MAIN_TITLE]
                file_name = ds[FILEPATH]
                file_path = os.path.join(self.data_directory, file_name)

                try:
                    df = None
                    try:
                        df = pd.read_csv(str(file_path))
                    except FileNotFoundError:
                        error_messages.append(f"ERROR: File not found for '{ds_title}': {file_path}")
                        continue
                    except UnicodeDecodeError:
                        try:
                            df = pd.read_csv(str(file_path), encoding='latin-1')
                            error_messages.append(f"WARNING: Read '{ds_title}' with latin-1 encoding.")
                        except Exception as e_enc:
                            error_messages.append(
                                f"ERROR: Failed to read '{ds_title}' with multiple encodings: {e_enc}")
                            continue
                    except pd.errors.EmptyDataError:
                        error_messages.append(f"WARNING: File for '{ds_title}' is empty or unparseable.")
                        continue
                    except Exception as e_read:
                        error_messages.append(f"ERROR: Reading file for '{ds_title}': {e_read}")
                        continue

                    if df is None:
                        continue

                    df.columns = df.columns.str.strip()

                    if not ('Variable' in df.columns and 'Description' in df.columns):
                        error_messages.append(
                            f"WARNING: Dataset '{ds_title}' skipped - missing 'Variable' or 'Description' column.")
                        continue

                    temp_df = df[['Variable', 'Description']].copy()
                    temp_df.columns = ['Variable Name', 'Description']

                    try:
                        temp_df['Variable Name'] = temp_df['Variable Name'].astype(str).fillna('')
                        temp_df['Description'] = temp_df['Description'].astype(str).fillna('')
                    except Exception as e_conv:
                        error_messages.append(
                            f"WARNING: Could not convert columns to string for '{ds_title}': {e_conv}. Skipping filtering for this file.")
                        continue


                    if not search_words:

                        filtered_df = temp_df
                    else:
                        final_mask = pd.Series([True] * len(temp_df), index=temp_df.index)

                        for word in search_words:
                            escaped_word = re.escape(word)

                            word_mask = (
                                    temp_df['Variable Name'].str.contains(escaped_word, case=False, na=False,
                                                                          regex=True) |
                                    temp_df['Description'].str.contains(escaped_word, case=False, na=False, regex=True)
                            )
                            final_mask &= word_mask

                        filtered_df = temp_df[final_mask]

                    if not filtered_df.empty:
                        filtered_df.insert(0, 'Dataset', ds_title)
                        data_list.append(filtered_df)

                except Exception as e_proc:
                    error_messages.append(f"ERROR: Processing dataset '{ds_title}': {e_proc}")

            with self.error_output:
                if error_messages:
                    print("Encountered issues while searching:")
                    for msg in error_messages:
                        print(f"- {msg}")

            if data_list:
                try:
                    combined_df = pd.concat(data_list, ignore_index=True)
                    if not combined_df.empty:
                        var_html = self.generate_combined_table(combined_df)
                        self.table_display.value = var_html
                    else:
                        self.table_display.value = "<p>No matching records found after combining.</p>"

                except Exception as e_concat:
                    self.table_display.value = ""
                    with self.error_output:
                        print(f"FATAL ERROR: Failed to combine results: {e_concat}")
            else:
                if not error_messages:
                    self.table_display.value = "<p>No matching records found.</p>"
                else:
                    self.table_display.value = "<p>Search incomplete due to errors. No matching records found in successfully processed files.</p>"


        else:  # Specific dataset selected
            file_info = self.dataset_info.loc[
                self.dataset_info[MAIN_TITLE] == self.dataset_select.value, FILEPATH]  # Use constants

            if not file_info.empty:
                file_name = file_info.values[0]
                file_path = os.path.join(self.data_directory, file_name)
                ds_title = self.dataset_select.value

                try:
                    df = None
                    try:
                        df = pd.read_csv(str(file_path))
                    except FileNotFoundError:
                        error_messages.append(f"ERROR: File not found: {file_path}")
                    except UnicodeDecodeError:
                        try:
                            df = pd.read_csv(str(file_path), encoding='latin-1')
                            error_messages.append(f"WARNING: Read file with latin-1 encoding.")
                        except Exception as e_enc:
                            error_messages.append(f"ERROR: Failed to read file with multiple encodings: {e_enc}")
                    except pd.errors.EmptyDataError:
                        error_messages.append(f"WARNING: File is empty or unparseable.")
                    except Exception as e_read:
                        error_messages.append(f"ERROR: Reading file: {e_read}")

                    if df is None:
                        self.table_display.value = ""

                    elif not ('Variable' in df.columns and 'Description' in df.columns):
                        error_messages.append(
                            f"ERROR: Dataset '{ds_title}' missing 'Variable' or 'Description' column.")
                        self.table_display.value = ""  # Clear table

                    else:
                        temp_df = df[['Variable', 'Description']].copy()
                        temp_df.columns = ['Variable Name', 'Description']

                        try:
                            temp_df['Variable Name'] = temp_df['Variable Name'].astype(str).fillna('')
                            temp_df['Description'] = temp_df['Description'].astype(str).fillna('')
                        except Exception as e_conv:
                            error_messages.append(
                                f"WARNING: Could not convert columns to string: {e_conv}. Filtering might be affected.")

                        if not search_words:
                            filtered = temp_df
                        else:
                            final_mask = pd.Series([True] * len(temp_df), index=temp_df.index)

                            for word in search_words:
                                escaped_word = re.escape(word)

                                word_mask = (
                                        temp_df['Variable Name'].str.contains(escaped_word, case=False, na=False,
                                                                              regex=True) |
                                        temp_df['Description'].str.contains(escaped_word, case=False, na=False,
                                                                            regex=True)
                                )

                                final_mask &= word_mask

                            filtered = temp_df[final_mask]

                        if not filtered.empty:
                            var_html = self.generate_variable_table(filtered)
                            self.table_display.value = var_html
                        else:
                            if not search_words:  # Should not happen if logic is correct unless df was empty
                                self.table_display.value = "<p>Dataset is empty.</p>"
                            else:
                                self.table_display.value = "<p>No matching records found in this dataset for the specified terms.</p>"

                except Exception as e_proc:
                    error_messages.append(f"ERROR: Processing dataset '{ds_title}': {e_proc}")
                    self.table_display.value = ""  # Clear display on processing error
            else:
                error_messages.append(f"ERROR: Dataset '{self.dataset_select.value}' not found in configuration.")
                self.table_display.value = ""

            with self.error_output:
                if error_messages:
                    print(f"Encountered issues while processing '{self.dataset_select.value}':")
                    for msg in error_messages:
                        print(f"- {msg}")


    def on_dataset_change(self, change):
        if change['type'] == 'change' and change['name'] == 'value':
            self.update_table(self.search_area.value)

    def on_search_clicked(self, b):
        self.table_display.value = ""
        self.error_output.clear_output()
        self.update_table(self.search_area.value)

    def save_filtered_table(self, _):
        self.error_output.clear_output()
        try:
            if not self.table_display.value.strip():
                with self.save_output:
                    self.save_output.clear_output()
                    print("No data to save")
            else:
                if self.dataset_select.value == "None":
                    filename = f"Datasets_with_{self.search_area.value}.html"
                else:
                    filename = f"{self.dataset_select.value}_{self.search_area.value}.html"
                with open(filename, "w", encoding="utf-8") as file:
                    file.write(self.table_display.value)

                with self.save_output:
                    self.save_output.clear_output()
                    print(f"Table saved as '{filename}'.")

        except Exception as e:
            with self.save_output:
                self.save_output.clear_output()
                print(f"Error saving file: {e}")
                # print("")

    def clear_output(self, b):
        self.table_display.value = ""
        self.save_output.clear_output()
        self.error_output.clear_output()
        self.dataset_select.value = 'None'
        self.search_area.value = ''




import ipywidgets as wd

from .styles import get_styles, style_base


def get_layout(box_name):
    layout_base = wd.Layout(grid_area=f'{box_name}', width='100%',
                            background_color='white', border='1px solid #ababab')
    return layout_base


def create_select_dropdown(options=None, box_name=None, value=None):
    select_dropdown = wd.Select(
        options=options,
        disabled=False,
        rows=10,
        value=value,
        layout=wd.Layout(grid_area=f'{box_name}_select_box', width='100%',
                         background_color='white', border='1px solid #ababab'))
    return select_dropdown


def create_multiple_select_dropdown(options=None, box_name=None):
    select_dropdown = wd.SelectMultiple(
        options=options,
        disabled=False,
        rows=10,
        layout=wd.Layout(grid_area=f'{box_name}_select_box', width='100%',
                         background_color='white', border='1px solid #ababab'))
    return select_dropdown


def create_dropdown(name, options=None, value=None):
    new_dropdown = wd.Dropdown(
        options=options,
        value=value,
        disabled=False,
        layout=wd.Layout(grid_area=f'{name}_dropdown_box')
    )
    return new_dropdown


def create_label(text, writing_style=None):
    if writing_style is None:
        writing_style = get_styles().get('label', '')
    lower_text = text.lower()
    label = wd.HTML(value=f"""<div style="{writing_style}"><p>Select {text}</p></div>""",
                    disabled=False,
                    layout=wd.Layout(grid_area=f'{lower_text}_label_box', width='100%',
                                     background_color='white', border='1px solid #ababab'),
                    style={**style_base, 'border': '1px solid #ababab'})
    return label


def create_helper(text, helper_name, writing_style=None):
    if writing_style is None:
        writing_style = get_styles().get('helper', '')
    helper = wd.HTML(value=f"""<div style="{writing_style}"><p>{text}</p></div>""",
                     disabled=False,
                     layout=wd.Layout(grid_area=f'{helper_name}_helper_box', border='1px solid gray', width='96%',
                                      justify='center'),
                     style={**style_base, 'word_break': 'break_all', 'padding': '3px'})
    return helper


def create_button(text, box_name, style):
    button = wd.Button(description=f'{text}', style=style,
                       layout=wd.Layout(grid_area=f'{box_name}_button_box', width='100%'))
    return button


def get_text_fields(text):
    new_field = wd.Text(value='', description=f'{text}: ',
                        style=dict(font_size='12px', text_color='black', background='#FFFFFF',
                                   description_width='150px'),
                        layout=wd.Layout(width='95%', height='50px', justify_content="flex-end"))
    return new_field


def get_date_picker(text):
    new_field = wd.DatePicker( description=f'{text}: ',
                              style=dict(font_size='12px', text_color='black', background='#FFFFFF',
                                         description_width='150px'),
                              layout=wd.Layout(width='95%', height='50px', justify_content="flex-end"))
    return new_field


def get_text_area(text):
    new_field = wd.Textarea(value='', description=f'{text}: ',
                            style=dict(font_size='12px', text_color='black', background='#FFFFFF',
                                       description_width='150px'),
                            layout=wd.Layout(width='95%', height='50px', justify_content="flex-end"))
    return new_field


def get_variable_text_area(text):
    lower_text = text.lower()
    new_text_area = wd.Textarea(value='', description=f'{text}',
                                placeholder=f'Enter column names for {lower_text}',
                                disabled=False, style=dict(description_width='150px'), layout=dict(width='700px'))
    return new_text_area


def button_styles(button_color):
    styles = dict(padding_block='10px', font_size='14px', button_color=f'{button_color}', text_color='white',
                  font_weight='bold', font_family=' Tahoma, Verdana,sans-serif')
    return styles


def get_label_buttons(text, color):
    new_button = wd.Button(description=f'{text}', style=button_styles(color),
                           layout=wd.Layout(margin='auto', display='block', width='fit-content', height='fit-content',
                                            padding_block='10px'))
    return new_button


def get_uploader(text):
    new_uploader = wd.FileUpload(accept='.csv', multiple=False, description=f'{text}',
                                 tooltip='Upload a CSV file', style=button_styles('firebrick'),
                                 layout=wd.Layout(margin='auto', display='block', width='fit-content',
                                                  height='fit-content', padding_block='10px'))
    return new_uploader




import os
import ipywidgets as wd
import pandas as pd
import re
from IPython.display import display, clear_output, HTML

from pyschare.helpers.constants import search_helper_text, MAIN_TITLE, FILEPATH, DATAPATH, DICTPATH
from pyschare.helpers.create_widgets import create_helper, create_select_dropdown, create_button
from pyschare.helpers.styles import get_styles, button_style
from pyschare.helpers.data_functions import get_data_dir,get_main_table_info, get_dictionary_table_info, get_visual_table_info


class _Search:
    def __init__(self):
        # self.dir_path = os.path.dirname(__file__)
        # self.csv_path = get_main_table_path()
        # self.dataset_info = get_search_table_info()
        self.dataset_info = get_main_table_info()
        self.data_directory = get_data_dir()
        self.search_helper = create_helper(text=search_helper_text, helper_name='search')
        self.styles = get_styles()

        self.dataset_select = create_select_dropdown(options=["None"] + self.dataset_info[MAIN_TITLE].tolist(),
                                                     box_name='dataset', value='None')

        self.search_area = wd.Textarea(value='', placeholder='Type at least 3 characters',
                                       layout=wd.Layout(grid_area='search_text_box', width='90%',
                                                        background_color='white', border='1px solid #ababab'),
                                       disabled=False)

        self.search_button = create_button(text='Search', box_name='search',
                                           style={**button_style, 'button_color': 'blue'})
        self.save_button = create_button(text='Save Table', box_name='save',
                                         style={**button_style, 'button_color': 'green'})
        self.clear_button = create_button(text='Clear Table', box_name='clear',
                                          style={**button_style, 'button_color': 'red'})

        self.dataset_label = wd.Label(value='Datasets', layout=wd.Layout(grid_area='dataset_label_box'))
        self.search_label = wd.Label(value='Search', layout=wd.Layout(grid_area='search_label_box'))
        self.table_display = wd.HTML(value="")
        self.error_output = wd.Output()

        self.save_button.on_click(self.save_filtered_table)
        self.clear_button.on_click(self.clear_output)
        self.search_button.on_click(self.on_search_clicked)
        self.save_output = wd.Output(layout=wd.Layout(width='100%'))
        self.dataset_select.observe(self.on_dataset_change, 'value')

        self.error_output.layout.grid_area = 'error_output_box'

        self.search_grid_layout = wd.GridBox(
            children=[self.search_helper, self.dataset_label, self.dataset_select,
                      self.search_area, self.search_label, self.search_button, self.save_output, self.save_button,
                      self.clear_button, self.error_output],
            layout=wd.Layout(grid_template_columns='40% 15% 30% 15%',
                             grid_template_rows='auto',
                             display='grid',
                             grid_template_areas='''
                                     "search_helper_box search_helper_box  search_helper_box search_helper_box"
                                    "dataset_label_box search_label_box . ."
                                     "dataset_select_box  search_text_box  search_text_box search_text_box "
                                    "dataset_select_box  . search_button_box ."
                                     "dataset_select_box  . save_button_box ."
                                      "dataset_select_box  . clear_button_box ."
                                       "dataset_select_box  . error_output_box ."
                                      "save_output_box  save_output_box save_output_box save_output_box"
                                    ''',
                             grid_gap='5px',
                             width='98%'
                             ))
        display(wd.VBox([self.search_grid_layout, self.table_display]))

    def generate_variable_table(self, df):
        html_content = f"""<html> <table style="{self.styles['table']}">
            <tr><th style="{self.styles['first_column']}">Variables</th>
            <th style="{self.styles['second_column']}">Descriptions</th>
            </tr>
        """
        for _, row in df.iterrows():
            html_content += f"""<tr><td style="{self.styles['first_cell']}">{row.iloc[0]}</td>
                <td style="{self.styles['second_cell']}">{row.iloc[1]}</td></tr>
            """
        html_content += "</table></html>"

        with open("variables.html", 'w') as file:
            file.write(html_content)
        return html_content

    def generate_combined_table(self, df):
        html_content = f"""
        <html><table style="{self.styles['table']}">
            <tr><th style="{self.styles['first_column']}">Dataset</th>
            <th style="{self.styles['second_column']}">Variables</th>
            <th style="{self.styles['third_column']}">Description</th>
            </tr>"""

        for _, row in df.iterrows():
            html_content += f"""
            <tr><td style="{self.styles['first_cell']}">{row.iloc[0]}</td>
            <td style="{self.styles['first_cell']}">{row.iloc[1]}</td>
            <td style="{self.styles['second_cell']}">{row.iloc[2]}</td>
            </tr>"""
        html_content += "</table></html>"

        with open("combined_table.html", 'w') as file:
            file.write(html_content)
        return html_content

    def update_table(self, search_text):
        pattern = ".*" + re.escape(search_text) + ".*" if search_text else ".*"

        if self.dataset_select.value == "None":
            data_list = []
            for idx, ds in self.dataset_info.iterrows():
                ds_title = str(ds[MAIN_TITLE])
                file_name = str(ds[FILEPATH])
                file_path = os.path.join(self.data_directory, file_name)

                try:
                    df = pd.read_csv(file_path)
                    if 'Variable' in df.columns and 'Description' in df.columns:
                        temp_df = df[['Variable', 'Description']]
                        temp_df.columns = ['Variable Name', 'Description']

                        filtered_df = temp_df[
                            temp_df['Variable Name'].str.contains(pattern, case=False, na=False, regex=True) |
                            temp_df['Description'].str.contains(pattern, case=False, na=False, regex=True)
                            ]
                        if not filtered_df.empty:
                            filtered_df.insert(0, 'Dataset', ds_title)
                            data_list.append(filtered_df)
                    else:
                        print(f"Dataset {ds_title} does not have required columns.")
                        # print("")
                except Exception as e:
                    print(f"Error loading dataset {ds_title}: {e}")
                    # print("")
            if data_list:
                combined_df = pd.concat(data_list, ignore_index=True)
            else:
                combined_df = pd.DataFrame(columns=['Dataset', 'Variable Name', 'Description'])

            var_html = self.generate_combined_table(combined_df)
            self.table_display.value = var_html
        else:
            file_info = self.dataset_info.loc[
                self.dataset_info[MAIN_TITLE] == self.dataset_select.value, FILEPATH]
            if not file_info.empty:
                file_name = file_info.values[0]
                file_path = os.path.join(self.data_directory, file_name)

                try:
                    df = pd.read_csv(str(file_path))
                    if 'Variable' in df.columns and 'Description' in df.columns:
                        temp_df = df[['Variable', 'Description']].copy()
                        temp_df.columns = ['Variable Name', 'Description']

                        filtered = temp_df[
                            temp_df['Variable Name'].str.contains(pattern, case=False, na=False, regex=True) |
                            temp_df['Description'].str.contains(pattern, case=False, na=False, regex=True)
                            ]
                        if not filtered.empty:
                            var_html = self.generate_variable_table(filtered)
                            self.table_display.value = var_html
                        else:
                            self.table_display.value = "No matching records found."
                            # self.table_display.value = ""

                    else:
                        self.table_display.value = "Error: Dataset does not have required columns."
                        # self.table_display.value = ""
                except Exception as e:
                    self.table_display.value = f"Error loading dataset: {e}"
                    # self.table_display.value = ""
            else:
                self.table_display.value = "Error: Dataset not found in MainTableDatasets.csv."
                # self.table_display.value = ""

    def on_dataset_change(self, change):
        if change['type'] == 'change' and change['name'] == 'value':
            self.update_table(self.search_area.value)

    def on_search_clicked(self, b):
        self.error_output.clear_output()
        self.update_table(self.search_area.value)

    def save_filtered_table(self, _):
        try:
            if not self.table_display.value.strip():
                with self.save_output:
                    self.save_output.clear_output()
                    print("No data to save")
            else:
                if self.dataset_select.value == "None":
                    filename = f"Datasets_with_{self.search_area.value}.html"
                else:
                    filename = f"{self.dataset_select.value}_{self.search_area.value}.html"
                with open(filename, "w", encoding="utf-8") as file:
                    file.write(self.table_display.value)

                with self.save_output:
                    self.save_output.clear_output()
                    print(f"Table saved as '{filename}'.")

        except Exception as e:
            with self.save_output:
                self.save_output.clear_output()
                print(f"Error saving file: {e}")
                # print("")

    def clear_output(self, b):
        self.table_display.value = ""
        self.save_output.clear_output()
        self.error_output.clear_output()
        self.dataset_select.value = 'None'
        self.search_area.value = ''








import os
import pandas as pd
import re
import warnings
import ipywidgets as wd

warnings.filterwarnings("ignore")
from pyschare.helpers.constants import MAIN_TITLE, FILEPATH

from pyschare.helpers.styles import get_styles
from pyschare.helpers.data_functions import get_data_dir, get_main_table_info



styles = get_styles()


def generate_variable_table(df):
    html_content = f"""<html> <table style="{styles['table']}">
        <tr><th style="{styles['first_column']}">Variables</th>
        <th style="{styles['second_column']}">Descriptions</th>
        </tr>
    """
    for _, row in df.iterrows():
        html_content += f"""<tr><td style="{styles['first_cell']}">{row.iloc[0]}</td>
            <td style="{styles['second_cell']}">{row.iloc[1]}</td></tr>
        """
    html_content += "</table></html>"
    return html_content


def generate_combined_table(df):
    html_content = f"""
    <html><table style="{styles['table']}">
        <tr><th style="{styles['first_column']}">Dataset</th>
        <th style="{styles['second_column']}">Variables</th>
        <th style="{styles['third_column']}">Description</th>
        </tr>"""

    for _, row in df.iterrows():
        html_content += f"""
        <tr><td style="{styles['first_cell']}">{row.iloc[0]}</td>
        <td style="{styles['first_cell']}">{row.iloc[1]}</td>
        <td style="{styles['second_cell']}">{row.iloc[2]}</td>
        </tr>"""
    html_content += "</table></html>"
    return html_content


class _DataSearch:
    def __init__(self):
        self.dataset_info = get_main_table_info()
        self.data_directory = get_data_dir()

    def update_table(self, search_text):
        error_messages = []
        search_words = search_text.strip().split() if search_text and search_text.strip() else []

        data_list = []
        for idx, ds in self.dataset_info.iterrows():
            ds_title = ds[MAIN_TITLE]
            file_name = ds[FILEPATH]
            file_path = os.path.join(self.data_directory, file_name)

            try:
                df = None
                try:
                    df = pd.read_csv(str(file_path))
                except FileNotFoundError:
                    error_messages.append(f"ERROR: File not found for '{ds_title}': {file_path}")
                    continue
                except UnicodeDecodeError:
                    try:
                        df = pd.read_csv(str(file_path), encoding='latin-1')
                        error_messages.append(f"INFO: Read '{ds_title}' with latin-1 encoding.")  # Changed to INFO
                    except Exception as e_enc:
                        error_messages.append(
                            f"ERROR: Failed to read '{ds_title}' with multiple encodings: {e_enc}")
                        continue
                except pd.errors.EmptyDataError:
                    error_messages.append(f"WARNING: File for '{ds_title}' is empty or unparseable.")
                    continue
                except Exception as e_read:
                    error_messages.append(f"ERROR: Reading file for '{ds_title}': {e_read}")
                    continue

                if df is None or df.empty:
                    continue

                df.columns = df.columns.str.strip()

                variable_col = 'Variable'
                description_col = 'Description'
                if not (variable_col in df.columns and description_col in df.columns):
                    found_var = False
                    found_desc = False
                    for col in df.columns:
                        if col.lower() == 'variable':
                            variable_col = col
                            found_var = True
                        if col.lower() == 'description':
                            description_col = col
                            found_desc = True
                    if not (found_var and found_desc):
                        error_messages.append(
                            f"WARNING: Dataset '{ds_title}' skipped - missing required columns (looked for '{variable_col}', '{description_col}').")
                        continue

                temp_df = df[[variable_col, description_col]].copy()
                temp_df.columns = ['Variable Name', 'Description']  # Standard names

                try:
                    temp_df['Variable Name'] = temp_df['Variable Name'].astype(str).fillna('')
                    temp_df['Description'] = temp_df['Description'].astype(str).fillna('')
                except Exception as e_conv:
                    error_messages.append(
                        f"WARNING: Could not convert columns to string for '{ds_title}': {e_conv}. Skipping filtering for this file.")
                    continue

                if not search_words:
                    filtered_df = temp_df
                else:
                    final_mask = pd.Series([True] * len(temp_df), index=temp_df.index)
                    for word in search_words:
                        escaped_word = re.escape(word)
                        try:
                            word_mask = (
                                    temp_df['Variable Name'].str.contains(escaped_word, case=False, na=False,
                                                                          regex=True) |
                                    temp_df['Description'].str.contains(escaped_word, case=False, na=False, regex=True)
                            )
                            final_mask &= word_mask
                        except Exception as e_filter:
                            error_messages.append(
                                f"WARNING: Error during filtering for word '{word}' in '{ds_title}': {e_filter}. Skipping word.")

                    filtered_df = temp_df[final_mask]

                if not filtered_df.empty:
                    filtered_df.insert(0, 'Dataset', ds_title)
                    data_list.append(filtered_df)

            except Exception as e_proc:
                error_messages.append(f"ERROR: Processing dataset '{ds_title}': {e_proc}")

        combined_df_for_term = pd.DataFrame()
        if data_list:
            try:
                combined_df_for_term = pd.concat(data_list, ignore_index=True)
            except Exception as e_concat:
                error_messages.append(f"ERROR: Failed to combine results for search term '{search_text}': {e_concat}")

        return combined_df_for_term, error_messages



VISUAL_TABLE: str = 'MainTableDatasetsDictionary_Visuals.csv'
MAIN_TABLE: str = 'MainTableDatasetsDictionary_2.csv'
MAIN_TABLE_VAR: str = 'dataset'
MAIN_TITLE: str = 'dataset_title'
DPATH: str = 'datasets'
FPATH: str = 'file_name'

DICTPATH: str = 'Dictionary_path'


visual_helper_text = "<p>Use the dropdown menus to select a dataset and configure your plot parameters.</p><ul><li><strong>Bar, count, box, boxen, strip, swarm, and violin plots</strong> typically require a categorical variable on the X-axis (or hue) and a numeric variable on the Y-axis; see the <a href='https://seaborn.pydata.org/tutorial/categorical.html' target='_blank'>categorical tutorial</a> for details.</li><li><strong>Scatter and line plots</strong> call for numeric variables on both axes (e.g., time vs. measurement); refer to the <a href='https://seaborn.pydata.org/tutorial/relational.html' target='_blank'> relational tutorial</a>.</li><li><strong>Histograms</strong> typically need a single numeric variable on the X-axis and are described in the <a href='https://seaborn.pydata.org/tutorial/distributions.html' target='_blank'> distributions tutorial</a>.</li></ul><p>Use <strong>“hue”</strong> to differentiate categories by color, <strong>“style”</strong> to vary markers or lines, and <strong>“size”</strong> to scale markers based on another variable. The <strong>“col” and “row”</strong> options create subplots (facets) for comparison across categories, while the <strong>“multiple”</strong> parameter (e.g., 'dodge,' 'stack,' 'fill') manages overlapping data displays. Once the plot type and settings are selected, click <strong>“Show Plot”</strong> to visualize the results.</p>"
select_helper_text = "<p>To view the dataset, use the <strong>Select Dataset</strong> dropdown and click the <strong>Show Data</strong> button. To save the displayed data, click the <strong>Save Data</strong> button. The confirmation that where data is saved will be shown below the Save Data button.</p><p> To clear both the confirmation message and the displayed data table, click the <strong>Clear Output</strong> button.</p>"

subset_helper_text = "<p>Use the <strong>Select Dataset</strong> dropdown to choose a dataset. The available variables will be dynamically populated when you select options in the <strong>Select Variables</strong> dropdown. After selecting the desired variables from the <strong>Select Variables</strong> dropdown, you may visualize the data by clicking the <strong>Show Data</strong> button. This will display the first few rows of the specific columns selected in the output area below.</p>     <p>To save the displayed data, click the <strong>Save Data</strong> button. This action will store the selected data in your bucket and confirm the successful operation in the output area. Please make sure you have made selections in both the dataset and variables dropdowns before attempting to save.</p>"

calculate_helper_text = f"""<p>This widget allows you to create various types of plots using your selected dataset. Follow the steps below to build your visualization:</p>

    <ol><li><strong>Choose a Dataset:</strong>
            <p>Begin by selecting a dataset from the "Select Dataset" dropdown menu. Click the dropdown to see the list of available datasets and choose the one you want to use.
            </p>
        </li><li>
            <strong>Select a Plot Type:</strong>
            <p>Next, choose the type of plot you want to create from the "Select Plot" dropdown menu. Common plot types include:
            </p>
            <ul>
                <li><strong>Bar Plots, Count Plots, Box Plots, Boxen Plots, Strip Plots, Swarm Plots, and Violin Plots:</strong> These are typically used to show relationships between categories. They usually require a categorical variable for the X-axis (or "Hue") and a numeric variable for the Y-axis. (See the <a href="link_to_categorical_tutorial">Categorical Tutorial</a> for more details).</li>
                <li><strong>Scatter Plots and Line Plots:</strong> These are used to show relationships between two numeric variables. For example, you might plot time versus measurement. (See the <a href="link_to_relational_tutorial">Relational Tutorial</a> for more details).</li>
                <li><strong>Histograms:</strong> These are used to show the distribution of a single numeric variable. (See the <a href="link_to_distributions_tutorial">Distributions Tutorial</a> for more details).</li>
            </ul>
        </li>

        <li>
            <strong>Configure Plot Parameters:</strong>
            <p>
                <ul>
                    <li><strong>X-Axis and Y-Axis:</strong> Use the "Select X" and "Select Y" dropdown menus to choose the variables you want to plot on the X and Y axes. The available options will depend on the dataset you selected.</li>
                    <li><strong>Hue:</strong> Use the "Select Hue" dropdown to color-code your data points based on categories. This helps to visualize how different categories are distributed.</li>
                    <li><strong>Style:</strong> Use the "Select Style" dropdown to vary the markers or lines in your plot, based on categories.</li>
                    <li><strong>Size:</strong> Use the "Select Size" dropdown to scale the size of the markers based on another variable.</li>
                    <li><strong>Column and Row:</strong> Use the "Select Column" and "Select Row" dropdowns to create subplots (facets). This allows you to compare different categories across multiple plots.</li>
                    <li><strong>Layer (Multiple):</strong> Use the "Select Layer" dropdown to manage how overlapping data points are displayed. Options like "Dodge," "Stack," and "Fill" are available.</li>
                </ul>
            </p>
        </li>

        <li>
            <strong>View Your Plot:</strong>
            <p>
                Once you have selected your plot type and configured the parameters, click the "Show Plot" button. Your plot will be displayed below the widget.
            </p>
        </li>

        <li>
            <strong>Clear Output:</strong>
            <p>
                To clear the displayed plot, click the "Clear Output" button.
            </p>
        </li>
    </ol>
    """

data_explore_helper_text = f"""<p>This widget allows you to explore and manipulate datasets. Follow the steps below to work with the data:</p>
    <ol><li><strong>Selecting a Dataset:</strong>
            <p>Use the "Select Dataset" dropdown menu to choose the dataset you want to work with. Click on the dropdown to see a list of available datasets and select your choice.
            </p></li><li><strong>Selecting Variables:</strong>
            <p>After selecting a dataset, the "Select Variable" dropdown menu will populate with a list of variables available within that dataset.
                Choose the variables from the "Select Variable" dropdown you want to analyze. You can select multiple variables from this dropdown.
                (<em>Note:</em> If you do not select any variables, actions will be applied to all variables).
            </p></li><li><strong>Viewing Data:</strong>
            <p>To view the first few rows of the selected dataset or the selected variables, click the "Show Data" button. The results will be displayed in the output area below the widget.
            </p></li>
        <li><strong>Describing Data:</strong>
            <p>To view summary statistics (like mean, median, standard deviation) for the selected variables, click the "Describe Data" button.
                The summary statistics will be shown in the output area below the widget.
                If you haven't selected any variables, statistics for all variables in the dataset will be displayed.
            </p></li><li><strong>Saving Data:</strong>
            <p>To save the displayed data (either the entire dataset or the subset of selected variables), click the "Save Data" button.
                A confirmation message, including the location where the data is saved, will be displayed in the output area below the buttons.
                <br>
                <em>Note:</em> Ensure you have selected a dataset and, if applicable, variables before clicking "Save Data."
            </p></li><li><strong>Clearing Output:</strong>
            <p>To clear both the confirmation message (from saving) and the displayed data table or statistics in the output area, click the "Clear Output" button.
            </p></li></ol>
"""


search_helper_text = f"""<p>This widget allows you to search for variables and descriptions across multiple datasets. Follow these instructions to effectively use the search functionality:
        <li><strong>If you want to search within a specific dataset:</strong> Use the "Datasets" dropdown menu to select the dataset you wish to search. Scroll down to see the list of available datasets and select your choice.
        </li><li><strong>If you want to search across all datasets:</strong>Leave the "Datasets" dropdown set to "None". This is the default option.
        </li>In the text box, type the word or phrase you want to search for and click the "Search" button. Note: You need to enter at least 3 characters for the search to function.</p>
        <p><strong>Search Results:</strong> The widget will display a table below the search box, showing the search results.
        <li><strong>If a specific dataset was selected:</strong> The table will show the variables and descriptions from that dataset that match your search terms.
        </li><li><strong>If "None" was selected:</strong> The table will show results from all datasets that match your search terms, including the dataset name, variable name, and description.
        </li></p><p><strong>Save Table:</strong> If you want to save the search results as an HTML file, click the "Save Table" button.
        <li><strong>If a specific dataset was selected:</strong> The file will be named using the dataset name and the search terms (e.g., "Food Security Data 2021_searchterm.html").
        </li><li><strong>If "None" was selected:</strong> The file will be named using "Datasets" and the search terms (e.g., "Datasets_with_searchterm.html").
        </li>A confirmation message will appear below the "Save Table" button, indicating the file name and location.</p>"""

datafacts_helper=f"""<p>Please provide the Project Title for this dataset. This title serves as the primary label for identification and citation purposes.
In the Project Description field, enter a detailed abstract or summary. This description should clearly outline the dataset's main objectives, scope, and content (such as key variables or topics covered). If applicable, include a brief summary of the methodology and potential areas for application or research.
Once done, press the 'Save' button to generate the dataset facts summary. </p>"""

metatable_helper=f"""<p>This section is designated for capturing comprehensive metadata – structured information that describes, explains, locates, or otherwise makes it easier to retrieve, use, or manage the dataset. Please complete the following fields with accurate and relevant information pertaining to your dataset.
</p><p> Providing detailed metadata is essential for ensuring the dataset adheres to principles of findability, accessibility, interoperability, and reusability (<em>FAIR</em>). Complete and accurate metadata significantly enhances the dataset's value by facilitating discovery, enabling proper interpretation and analysis, ensuring correct citation, and promoting responsible reuse within the community.
Key metadata elements often cover areas such as:
<li>
<strong>Identification:</strong> <code>Filename</code>, <code>URL</code> </li>
<li><strong>Format &amp; Structure:</strong> <code>Format</code>, <code>Rows</code>, <code>Columns</code>, <code>CDEs</code> (Common Data Elements), <code>Missing</code> data information</li>
<li><strong>Context &amp; Scope:</strong> <code>Domain</code>, <code>Keywords</code>, <code>Type</code>, <code>Geography</code>, <code>Description</code></li>
<li><strong>Methodology:</strong> <code>Data Collection Method</code>, <code>Time Method</code>, <code>Data Collection Timeline</code></li>
<li><strong>Administration &amp; Access:</strong> <code>License</code>, <code>Released</code> date, <code>Funding Agency</code></li>
</ul>
After completing the applicable fields, click the 'Save' button to commit the metadata and render it in a structured tabular format for review.</p>"""

provenance_helper=f"""<p>
  Please document the origin and authorship of this dataset by providing information about both the 
  <code>Source</code> and the <code>Author(s)</code>. For the Source, enter details about the entity responsible for the data's creation or 
  distribution (e.g., Institution Name, Project Name, Repository), including its <code>URL</code> 
  and a contact <code>Email</code> address if available. For the Author(s), include information about the individual creator(s) involved 
  (e.g., Name, Affiliation, ORCID iD).Once you have entered these details, click the <strong>"Save"</strong> button. This will record the 
  provenance data and generate a table summarizing the dataset's lineage for review.
</p>"""


dictionary_helper=f"""
    <p> A data dictionary provides essential definitions for each variable within your dataset 
        (e.g., variable names, detailed descriptions, units of measurement, coding schemes for categorical data).
  Before uploading, prepare your data dictionary as a CSV (Comma Separated Values) file. Ensure this file contains 
            at least two columns: one listing the exact <strong>variable names</strong> as they appear in your dataset, 
            and another providing their corresponding <strong>descriptions</strong> or definitions. Click the <strong>"Upload Data Dictionary"</strong> button and select your prepared CSV file from your system.
After uploading, you may be prompted to confirm which columns in your file correspond to the 
            'Variable Name' and 'Description' fields required by the system.
        Upon successful upload and mapping, a structured table detailing your dataset's variables and their 
        definitions will be generated based on the information in your dictionary file.
    </p>
"""


stats_helper=f"""
    <p>This section calculates and displays summary statistics for the variables in your dataset. To ensure 
        that appropriate statistical measures are computed for each variable, you must first classify the variables 
        according to their measurement scale (nominal, ordinal, discrete, continuous). To generate the summary statistics:
    <ol><li><strong>Upload/Confirm Dataset:</strong> Ensure your dataset (in CSV format) is loaded. You can use the 
            <strong>"Show Data"</strong> button to preview the data and verify variable names if needed.
        </li><li>
            <strong>Classify Variables by Type:</strong> Based on the nature of each variable in your dataset, input 
            its exact name into the corresponding classification field below. Separate multiple variable names within 
            a single category using commas (e.g., <code>Variable_a, Variable_b, Variable_c</code>).
            <ul>
                <li><strong>Ordinal Variables:</strong> Categorical variables with a meaningful order (e.g., education level: High School, Bachelor's, Master's).</li>
                <li><strong>Nominal Variables:</strong> Categorical variables without an inherent order (e.g., Country, Eye Color).</li>
                <li><strong>Continuous Variables:</strong> Numerical variables that can take any value within a given range (e.g., Temperature, Height).</li>
                <li><strong>Discrete Variables:</strong> Numerical variables that can only take specific, distinct values, often integers (e.g., Number of Children, Page Visits).</li>
            </ul></li>
        <li><strong>Generate Statistics:</strong> After assigning variables to their respective types, click the 
            <strong>"Show Statistics Table"</strong> button.
        </li></ol>
        The system will then generate and display a table containing relevant descriptive statistics tailored 
        to each variable's classification (e.g., frequency counts and percentages for nominal/ordinal; mean, median, 
        standard deviation, min/max for continuous/discrete).
    </p>
"""


bar_plot_helper = f"""
    <p>
        Bar plots display the frequency or count of observations for each category within your selected 
      categorical variables. This helps in understanding the distribution across different categories. Select one, two, or up to three categorical variables (nominal or ordinal) 
        from the dropdown menus provided. The corresponding bar plot(s) will automatically generate or update below as you make or change 
        your selections, illustrating the frequency for each category within the chosen variable(s).
    </p>
"""


plot_helper= f"""
    <p>To explore the relationship between any two variables from your dataset, choose two variables using the dropdown menus.   Pair plots that visualize the distributions and relationships between the chosen pair of variables will automatically generate or update in the space below as you make your selections. </p>"""

correlation_helper=f""" <p>A correlation heatmap provides a visual summary of the association matrix between selected variables. 
    Using color intensity, it illustrates the strength and direction of these associations, helping to identify patterns in your data.
Choose the variables you wish to include in the analysis from the provided lists. The heatmap automatically calculates and visualizes the appropriate measure of association based on the types of variables being compared:
<ul>
    <li>For pairs of <strong>continuous variables</strong>, it calculates <strong>Pearson's correlation coefficient (r)</strong> to measure linear association.</li>
    <li>For pairs of <strong>categorical variables</strong>, it calculates <strong>Cramér's V</strong> to measure the strength of association.</li>
    <li>For pairs involving one <strong>continuous</strong> and one <strong>categorical</strong> variable, it calculates <strong>Pearson's correlation coefficient (r)</strong>. <em>(Note: The categorical variable is automatically converted into numerical format using dummy coding for this calculation.)</em></li>
</ul>
The resulting heatmap will automatically generate or update below as you select or deselect variables.
</p>"""


metadata = {
    "Filename": "",
    "Format": "",
    "Study/Project URL": "",
    "Domain": "",
    "Keywords": "",
    "Type": "",
    "Geography": "",
    "Data Collection Method": "",
    "Time Method": "",
    "Rows": "",
    "Columns": "",
    "CDEs": "",
    "Missing": "",
    "License": "",
    "Released": (None),
    "Data Collection Timeline": (None, None),
    "Funding Agency":"",
    "Description": ""
}

output_names = ['facts', 'meta', 'pro', 'corr_plot', 'pair_plot', 'bars', 'stats', 'show_data', 'var']

button_labels = ['pro', 'meta', 'facts']

field_names = ['Project Title','Project Description', 'Filename', 'Format', 'URL','Domain','Keywords','Type', 'Geography' , 'Data Collection Method', 'Time Method', 'Rows',
               'Columns','CDEs', 'Missing','License', 'Released', 'Funding Agency', 'Description', 'Source Name', 'Source URL', 'Source Email', 'Author Name', 'Author URL', 'Author Email']







import ipywidgets as wd

def get_styles():
    styles = {
        "div": "padding-left: 10px; padding-right: 10px; margin: 20px auto;",
        "table": "width: 100%; border-spacing: 0; border-bottom: 1px solid black;",
        "title": "background-color: #FFFFFF; text-align: left; font-size: 20px; font-weight: bold; font-family: Helvetica, Neue; border-bottom: 3px solid black;",
        "first_title": "background-color: #FFFFFF; width: 30%;text-align: left; font-size: 20px; font-weight: bold; font-family: Helvetica, Neue; border-bottom: 3px solid black;",
        "cell": "background-color: #FFFFFF; text-align: right; border-spacing: 0; border-bottom: 1px solid black; font-family: Helvetica, Neue;word-break:break-all;",
        "stats_cell": "background-color: #FFFFFF; text-align: center; border-spacing: 0; border-bottom: 1px solid black; font-family: Helvetica, Neue;",
        "key_cell": "background-color: #FFFFFF; text-align: left; border-spacing: 0;font-weight: bold;  border-bottom: 1px solid black; font-family: Helvetica, Neue;",
        "key_cell_variable": "background-color: #FFFFFF; text-align: left; border-spacing: 0;font-weight: bold;  border-bottom: 1px solid black; font-family: Helvetica, Neue;  word-break:break-all;",
        "sub_cell": "background-color: #FFFFFF;text-align: left; border-spacing: 0;font-weight: bold;  border-bottom: 1px solid black; font-family: Helvetica, Neue;  padding-left: 16px",
        "first_header": "text-align: left; border-spacing: 0; border-bottom: 1px solid black; background-color: white; font-family: Helvetica, Neue; font-weight: bold",
        "header": "text-align: center; border-spacing: 0; border-bottom: 1px solid black; background-color: white; font-family: Helvetica, Neue; font-weight: bold",
        "stats_header": "text-align: center; border-spacing: 0; border-bottom: 1px solid black; background-color: white; font-family: Helvetica, Neue; font-weight: bold; word-break:break-all;",
        "label": "text-align: left; font-weight: bold; font-size: 12px; margin-left: 5px; font-family: Helvetica, Neue",
        "helper": "text-align: left; font-size: 14px; margin-left: 5px; margin-right: 5px; font-family: Helvetica, Neue",
        "first_column": "background-color: #FFFFFF; text-align: left; width: 25%; font-size: 12px; font-weight: bold; font-family: Tahoma, Verdana, serif; border-bottom: 3px solid black;",
        "second_column": "background-color: #FFFFFF; text-align: left; width: 25%; font-size: 12px; font-family: Tahoma, Verdana, serif; border-bottom: 3px solid black;",
        "third_column": "background-color: #FFFFFF; text-align: right; font-size: 12px; font-family: Tahoma, Verdana, serif; border-bottom: 3px solid black;",
        "first_cell": "background-color: #FFFFFF; text-align: left; border-spacing: 0; border-bottom: 1px solid black;font-size: 12px; font-family:Tahoma, Verdana; word-break:break-all;",
        "second_cell": "background-color: #FFFFFF; text-align: right; border-spacing: 0;  border-bottom: 1px solid black;font-size: 12px; font-family: Tahoma, Verdana; word-break:break-all;"
    }
    return styles


style_base = {'font_size': '14px', 'text_color': 'black', 'background': 'rgb(247, 247, 247)'}

button_style = {'font_size': '14px', 'text_color': 'white',
                'font_weight': 'bold', 'font_family': ' Tahoma, Verdana,sans-serif',
                'text_align': 'center'}

meta_layout = dict(width='95%', height='50px', justify_content="flex-end")
meta_style = dict(font_size='12px', text_color='black', background='#FFFFFF',
                  description_width='150px')

box_layout = wd.Layout(display='flex', flex_flow='column', align_items='center', width='100%')
vbox_layout = wd.Layout(padding_block='10px', margin = '2px')

def get_label_styles():
    styles = {
        "table": "width: 100%; border-spacing: 0; border-bottom: 1px solid black;",
        "title": "background-color: #FFFFFF; text-align: left; font-size: 20px; font-weight: bold; font-family: Tahoma, Verdana, serif; border-bottom: 3px solid black;",
        "first_title": "background-color: #FFFFFF; width: 30%;text-align: left; font-size: 20px; font-weight: bold; font-family: Tahoma, Verdana, serif; border-bottom: 3px solid black;",
        "cell": "background-color: #FFFFFF; text-align: right; border-spacing: 0; border-bottom: 1px solid black; font-family:Georgia, serif;word-break:break-all;",
        "stats_cell": "background-color: #FFFFFF; text-align: center; border-spacing: 0; border-bottom: 1px solid black; font-family:Georgia, serif;",
        "key_cell": "background-color: #FFFFFF; text-align: left; border-spacing: 0;font-weight: bold;  border-bottom: 1px solid black; font-family: Georgia, serif;",
        "key_cell_variable": "background-color: #FFFFFF; text-align: left; border-spacing: 0;font-weight: bold;  border-bottom: 1px solid black; font-family: Georgia, serif;  word-break:break-all;",
        "sub_cell": "background-color: #FFFFFF;text-align: left; border-spacing: 0;font-weight: bold;  border-bottom: 1px solid black; font-family: Georgia, serif;  padding-left: 16px",
        "first_header": "text-align: left; border-spacing: 0; border-bottom: 1px solid black; background-color: white; font-family: Georgia, serif; font-weight: bold",
        "header": "text-align: center; border-spacing: 0; border-bottom: 1px solid black; background-color: white; font-family: Georgia, serif; font-weight: bold",
        "stats_header": "text-align: center; border-spacing: 0; border-bottom: 1px solid black; background-color: white; font-family: Georgia, serif; font-weight: bold; word-break:break-all;",
    }
    return styles




import pandas as pd
import io
import os
import subprocess

from pyschare.helpers.constants import MAIN_TABLE, MAIN_TITLE, DICTPATH,VISUAL_TABLE, DATAPATH, FILEPATH

def get_main_table_path():
    dir_path = os.path.dirname(os.path.abspath(__file__))
    csv_path = os.path.join(dir_path, DATAPATH, MAIN_TABLE)
    return csv_path

def get_data_dir():
    dir_path = os.path.dirname(os.path.abspath(__file__))
    # csv_path = os.path.join(dir_path, DPATH, MAIN_TABLE)
    data_directory = os.path.join(dir_path, DATAPATH)
    return data_directory


def get_data(input_data):
    df = pd.read_csv(input_data)
    return df

def get_dictionary_table_info():
    dir_path: str = os.path.dirname(os.path.abspath(__file__))
    csv_path: str = os.path.join(dir_path, DATAPATH, MAIN_TABLE)
    dataset_info = pd.read_csv(csv_path)
    return dataset_info

def get_visual_table_info():
    dir_path: str = os.path.dirname(os.path.abspath(__file__))
    csv_path: str = os.path.join(dir_path, DATAPATH, VISUAL_TABLE)
    dataset_info = pd.read_csv(csv_path)
    return dataset_info

def get_main_table_info():
    dir_path: str = os.path.dirname(os.path.abspath(__file__))
    csv_path: str = os.path.join(dir_path, DATAPATH, MAIN_TABLE)
    dataset_info = pd.read_csv(csv_path)
    return dataset_info

def get_bucket():
    my_bucket = os.getenv('WORKSPACE_BUCKET')
    return my_bucket


def select_data_options(dataset_info) -> list:
    options = dataset_info[MAIN_TITLE].tolist()
    return options


def get_dropdown_value(dropdown):
    return dropdown.value

def get_input_path(selected_data):
    if selected_data is not None and not selected_data.empty:
        path = selected_data.iloc[0]['Data']
        dataset_name = selected_data.iloc[0][FILEPATH]
        return path, dataset_name
    else:
        print("No data found for the selected dataset title.")
        return None

def get_search_input_path(selected_data):
    if selected_data is not None and not selected_data.empty:
        path = selected_data.iloc[0]['Dictionary_path']
        dataset_name = selected_data.iloc[0][FILEPATH]
        return path, dataset_name
    else:
        print("No data found for the selected dataset title.")
        return None


def save_data_to_bucket(path, dataset_name):
    """
    :return: data_path not data
    """
    # path, dataset_name = get_input_path()
    my_bucket = get_bucket()
    args = ["gsutil", "cp", f"{path}", f"{my_bucket}/{dataset_name}"]
    output = subprocess.run(args, capture_output=True, text=True)
    if output.returncode == 0:
        target_path = f"{my_bucket}/{dataset_name}"
        return target_path


def save_to_bucket(dataframe, my_bucket):
    buffer = io.StringIO()
    dataframe.to_csv(buffer, index=False)
    buffer.seek(0)
    args = ['gsutil', 'cp', '-', my_bucket]
    process = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    process.communicate(buffer.getvalue().encode('utf-8'))


def load_data(path):
    """
    :return: data_path not data
    """
    # path, dataset_name = get_input_path()
    args = ["gsutil", "cat", f"{path}"]

    process = subprocess.Popen(args, stdout=subprocess.PIPE)
    output, _ = process.communicate()
    return output


def parse_data(dataset_name, output):
    if dataset_name.endswith("csv") or dataset_name.endswith("txt"):
        uploaded_data = pd.read_csv(io.StringIO(output.decode('utf-8')), low_memory=False)
        return uploaded_data

    if dataset_name.endswith("tsv"):
        uploaded_data = pd.read_csv(io.StringIO(output.decode('utf-8')), sep= '\t', low_memory=False)
        return uploaded_data

    elif dataset_name.endswith("xlsx"):
        uploaded_data = pd.read_excel(io.BytesIO(output))
        return uploaded_data

    # elif dataset_name.endswith("xpt"):
    #     uploaded_data = pd.read_sas(io.StringIO(output.decode('utf-8')), low_memory=False)
    #     return uploaded_data

    else:
        print("Unsupported file type.")
        return None


def get_parsed_data(selected_data):
    path, dataset_name = get_input_path(selected_data)
    output = load_data(path)
    parsed_data = parse_data(dataset_name, output)
    return parsed_data


